-- Copyright (C) 1995,2000,2002 Rational Software Corporation.
-- All rights reserved.

-- RESTRICTED RIGHTS LEGEND
--
-- Use, duplication, or disclosure by the Government and Boeing is subject to
-- restrictions as set forth in subparagraph (c)(1)(ii) of the Rights in
-- Technical Data and Computer Software clause of DFAR 252.227-7013 and in
-- similiar clauses in the FAR and NASA FAR Supplement.
------------------------------------------------------------------------------
-- PROPRIETARY AND CONFIDENTIAL INFORMATION OF RATIONAL SOFTWARE CORP.,
-- USE OR COPYING WITHOUT EXPRESS WRITTEN AUTHORIZATION IS STRICTLY PROHIBITED.
-- THIS MATERIAL IS PROTECTED AS AN UNPUBLISHED WORK UNDER THE U.S. COPYRIGHT
-- ACT OF 1976.  CREATED 1995.  ALL RIGHTS RESERVED.
------------------------------------------------------------------------------
--
-- File Name: tm_mcdc_allocation.1.ada
--
-- All memory allocations performed by the MCDC runtime support packages
-- are performed by this package.  This allows modification of the memory
-- allocation model for these packages without requiring changes throughout
-- these support packages.  This also allows monitoring code to be easily
-- added to track the amount of dynamic heap used by these packages.
--
-- Data can be allocated statically from a pool or dynamically on the global
-- heap. A data pool is a statically allocated array variable that is
-- statically initialized to (others => 0). The last element of a storage
-- pool is never used. So a client should allocate at least one extra storage
-- element.
--
-- The storage pool index type should be an integer type with 'First = 0.
-- The allocation routines attempt to allocate elements from the pool first.
-- Once the pool is exhausted elements are allocated on the global heap.
-- The Deallocate subprograms ignore objects that are allocated from the pool.
-- To deallocate all objects from a pool use the Clear subprogram

with Unchecked_Conversion;
with System;

package Tm_Mcdc_Allocation is
    pragma Not_Elaborated;

    generic
        type Element is private;
        type Handle is access Element;
    package Heap is

        function Allocate return Handle;
        function Allocate (Initial_Value : Element) return Handle;
        function Allocate_Array (Length : Integer) return Handle;

        procedure Deallocate (The_Handle : in out Handle);
        procedure Deallocate_Array
                     (The_Handle : in out Handle; Length : Integer);

        pragma Inline (Allocate, Allocate_Array, Deallocate, Deallocate_Array);
    end Heap;


    generic
        type Element is private;
        type Handle is access Element;
        type Pool_Index is range <>;
        type Pool_Storage is array (Pool_Index) of Element;
    package Pool is

        type Pool_Storage_Lock is range -128 .. 127;
        for Pool_Storage_Lock'Size use 8;
        type Pool_Storage_Locks is array (Pool_Index) of Pool_Storage_Lock;

        type Data is
            record
                Storage : Pool_Storage;
                Locks   : Pool_Storage_Locks;
                Next    : Pool_Index;
                --[ada95] pragma Volatile (Next);
            end record;
        pragma Convention (Ada_Pass_By_Reference, Data);

        type Data_Ptr is access Data;
        for Data_Ptr'Storage_Size use 0;

        function As_Data_Ptr is new Unchecked_Conversion
                                       (System.Address, Data_Ptr);

        function Allocate (From : Pool.Data_Ptr) return Handle;
        function Allocate (Initial_Value : Element; From : Pool.Data_Ptr)
                          return Handle;
        function Allocate_Array
                    (Length : Integer; From : Pool.Data_Ptr) return Handle;

        procedure Deallocate (The_Handle : in out Handle; From : Pool.Data_Ptr);
        procedure Deallocate_Array (The_Handle : in out Handle;
                                    Length     : Integer;
                                    From       : Pool.Data_Ptr);

        procedure Clear (Pool_Data : Pool.Data_Ptr);

        pragma Inline (Deallocate, Deallocate_Array);
    end Pool;

    --
    -- Several state variables are maintained to track the current allocation
    -- statistics for the MCDC support runtime.  Note, these values will not
    -- always be actively maintained.
    --
    Total_Alloc_Size   : Integer := 0;
    Total_Dealloc_Size : Integer := 0;
    Current_Alloc_Size : Integer := 0;
    Peak_Alloc_Size    : Integer := 0;
    Unfulfilled_Allocs : Integer := 0;

end Tm_Mcdc_Allocation;
