--+--------------------------------------------------------------------------+
--| IBM Confidential                                                         |
--| OCO Source Materials                                                     |
--| Rational Apex                                                            |
--| (C) Copyright IBM Corp. 2008 All Rights Reserved                         |
--|                                                                          |
--| The source code for this program is not published or otherwise divested  |
--| of its trade secrets, irrespective of what has been deposited with the   |
--+--------------------------------------------------------------------------+
--| Copyright 1992 - 2002 Rational Software Corporation. All Rights Reserved.|
--| This software contains proprietary and confidential information of       |
--| Rational and its suppliers. Use, disclosure or reproduction is           |
--| prohibited without the prior express written consent of Rational.        |
--+--------------------------------------------------------------------------+
--

with System;
with Tm_Mcdc_Runtime_Support;
with Unchecked_Conversion;
with Apex_Unix_Types;
with Apex_Unix_Operations;
with Apex_Unix_Constants;

package body Tm_Mcdc_Io_Support is

    package Mrs renames Tm_Mcdc_Runtime_Support;


    ---------------------------------------------------------------------------
    -- Support Routines
    ---------------------------------------------------------------------------

    function Id_Image (The_Id : Integer) return String is
    begin
        if The_Id < 0 then
            return "";
        else
            return '.' & Mrs.Mcdc_String_Utilities.Number_To_String
                            (Mrs.Trace_Integer (The_Id));
        end if;
    end Id_Image;

    function Environment_Value
                (Var : String; Default_Value : String) return String is

        subtype C_String is Apex_Unix_Types.Char_Ptr;
        type C_Array is array (Natural range 0 .. 32767) of C_String;
        type C_Array_Ptr is access C_Array;
        for C_Array_Ptr'Storage_Size use 0;

        function Convert is new Unchecked_Conversion
                                   (Source => Apex_Unix_Types.Char_Ptr_Ptr,
                                    Target => C_Array_Ptr);

        Environment : constant C_Array_Ptr :=
           Convert (Apex_Unix_Operations.Envp);
        Rover       : Natural              := 0;
    begin  
        loop
            declare
                Arg : constant Apex_Unix_Types.Char_Ptr := Environment (Rover);
            begin
                if Apex_Unix_Types."=" (Arg, null) then
                    -- The environment list is terminated by a null pointer;
                    -- if we hit it, return the default value for the pointer.
                    return Default_Value;
                end if;

                -- Look for a string of the form "VAR=VALUE"

                declare
                    Assoc  : constant String  :=
                       Apex_Unix_Types.To_String (Arg);
                    Eq_Pos : constant Integer := Assoc'First + Var'Length;
                begin
                    if Eq_Pos <= Assoc'Last and then
                       Assoc (Eq_Pos) = '=' and then
                       Var = Assoc (Assoc'First .. Eq_Pos - 1) then

                        return Assoc (Eq_Pos + 1 .. Assoc'Last);
                    end if;
                end;

                Rover := Rover + 1;
            end;
        end loop;
    end Environment_Value;

    function Trace_File_Name
                (The_Id : Integer; Use_Suffix : String) return String is
        --
        -- The trace file name is currently based on the name of the command
        -- being executed.  In the future we will likely want to only use the
        -- command name as a last resort and trigger on an environment variable
        -- when possible.  The environment variable will hopefully be the name
        -- of the test results file with a ".trace" suffix.
        --
        -- See if TestMate has passed a file name along ...
        The_Trace_File : constant String :=
           Environment_Value ("MCDC_TRACE_NAME", "");
        The_Start      : Natural         :=
           Mrs.Mcdc_String_Utilities.Reverse_Locate ('.', The_Trace_File);
        The_Path       : Natural         :=
           Mrs.Mcdc_String_Utilities.Reverse_Locate ('/', The_Trace_File);
    begin
        if The_Trace_File = "" then
            -- If TestMate did not specify a name then use the command name
            -- as a basis for the trace file name ...
            declare
                The_Command : constant String :=
                   Apex_Unix_Types.To_String (Apex_Unix_Operations.Argv.all);
            begin
                return The_Command & Id_Image (The_Id) & '.' & Use_Suffix;
            end;
        elsif The_Start = 0 or else The_Start < The_Path then
            return The_Trace_File & Id_Image (The_Id) & '.' & Use_Suffix;
        else -- just replace existing suffix with required suffix ...
            return The_Trace_File (The_Trace_File'First .. The_Start - 1) &
                      Id_Image (The_Id) & '.' & Use_Suffix;
        end if;
    end Trace_File_Name;


    function Trace_File_Name
                (The_Id : Integer; Is_Scorecard : Boolean) return String is
    begin
        -- For self-hosted system, derive the trace file name from an environment 
        -- variable
        if Is_Scorecard then
            return Trace_File_Name (The_Id, "score");
        else
            return Trace_File_Name (The_Id, "trace");
        end if;
    end Trace_File_Name;

    ---------------------------------------------------------------------------
    -- Interface Routines
    ---------------------------------------------------------------------------

    procedure Create (The_Obj      : in out Object;
                      The_Id       : Integer;
                      Is_Scorecard : Boolean) is
        The_C_Filename : constant String :=
           Trace_File_Name (The_Id, Is_Scorecard) & Ascii.Nul;
        use Apex_Unix_Constants.Unix_Fcntl;
        use Apex_Unix_Constants.Unix_Stat;

        The_Open_Mode : constant := S_Irusr + S_Iwusr + S_Irgrp + S_Iroth;
    begin
        The_Obj.The_File :=
           Integer (Apex_Unix_Operations.Open
                       (Path  =>
                           Apex_Unix_Types.To_Char_Ptr
                              (The_C_Filename (The_C_Filename'First)'Address),
                        Oflag => Apex_Unix_Types.Int
                                    (O_Wronly + O_Creat + O_Trunc),
                        Mode  => Apex_Unix_Types.Mode_T'(The_Open_Mode)));
    end Create;


    procedure Close (The_Obj : in out Object) is
        The_Status : Apex_Unix_Types.Int;
    begin
        if The_Obj.The_File /= 0 then
            The_Status       := Apex_Unix_Operations.Close
                                   (Apex_Unix_Types.Int (The_Obj.The_File));
            The_Obj.The_File := 0;
        end if;
    end Close;


    procedure Put_Line (The_Obj : Object; The_Line : String) is
        The_Size : Apex_Unix_Types.Ssize_T;
        Lf       : Character := Ascii.Lf;
    begin
        The_Size := Apex_Unix_Operations.Write
                       (Apex_Unix_Types.Int (The_Obj.The_File),
                        Apex_Unix_Types.To_Char_Ptr
                           (The_Line (The_Line'First)'Address),
                        Apex_Unix_Types.Size_T (The_Line'Length));
        The_Size := Apex_Unix_Operations.Write
                       (Apex_Unix_Types.Int (The_Obj.The_File),
                        Apex_Unix_Types.To_Char_Ptr (Lf'Address), 1);
    end Put_Line;

    procedure Put_Message (The_Obj  : Object;
                           The_Line : String;
                           The_Kind : Log_Msg_Kind := Error_Msg) is
        Prefix_Id : constant array (Log_Msg_Kind) of Character :=
           (Note_Msg => 'N', Warning_Msg => 'W', Error_Msg => 'E');
    begin
        Put_Line (The_Obj  => The_Obj,
                  The_Line => "%%" & Prefix_Id (The_Kind) & ' ' & The_Line);
    end Put_Message;

end Tm_Mcdc_Io_Support;
pragma Suppress_Elaboration_Checks;
