--+--------------------------------------------------------------------------+
--| Copyright 1992 - 2002 Rational Software Corporation. All Rights Reserved.|
--| This software contains proprietary and confidential information of       |
--| Rational and its suppliers. Use, disclosure or reproduction is           |
--| prohibited without the prior express written consent of Rational.        |
--+--------------------------------------------------------------------------+
--
------------------------------------------------------------------------------
--
--
-- Customization Notes:
--
--     This file defines the interface between the instrumented coverage
--     harness and explicitly called runtime services.  This file and 
--     the Tm_Mcdc_Config package are typically the only files that you
--     ever need to customize for the instrumented coverage harness.
-- 
--     For native (self-hosted), Rational Exec, Tornado, and LynxOS targets,
--     this file should not require any customization at all.
--
--
--    Exit Handler:
--       If this is not supported, it can be safely no-op'ed.  But, if
--       no-op'ed, the harness will then no longer work for program test
--       cases, only for Ada test cases.
--
--    Dynamic Memory Allocation:
--       This can be safely no-op'ed.  If you do so, make sure the buffers
--       specified in Tm_Mcdc_Config are sufficiently large.
--
--    String Utilities:
--       Don't change this -- you will only get disaster.
--

with Apex_Ada_Krn_Defs;
with Apex_Ada_Krn_I;

package body Tm_Mcdc_Runtime_Support is

    Base_Characters :
       constant array (Trace_Integer range -15 .. 15) of Character :=
       ('F', 'E', 'D', 'C', 'B', 'A', '9', '8', '7', '6', '5',
        '4', '3', '2', '1', '0', '1', '2', '3', '4', '5',
        '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F');


    procedure Register_Exit_Handler is
        -- This is used to guarantee that the harness will be called upon 
        -- program exit to output the coverage information.
        --
        -- There are three choices here.
        -- 1. Call Apex_Ada_Krn_I.Callout_Install
        --    This works for Rational runtimes.
        -- 2. Make a call to the runtime service, but outside the Ada
        --    closure.
        --    This is desireable for instrumenting runtimes that are 
        --    provided in source, such as MARK.  The drawback here is that
        --    there's no guarantee that the proper underlying code gets
        --    elaborated in time.
        -- 3. Stub it out
        --    This is not a problem for Ada test cases, but you won't get
        --    the data for program test cases.
        Dummy : Boolean;
    begin
        -- Call for any normal exit event.
        -- And call for any exceptional exit event.
        Dummy := Apex_Ada_Krn_I.Callout_Install
                    (Apex_Ada_Krn_Defs.Exit_Event, Exit_Handler'Address);
        Dummy := Apex_Ada_Krn_I.Callout_Install
                    (Apex_Ada_Krn_Defs.Unexpected_Exit_Event,
                     Exit_Handler'Address);
    end Register_Exit_Handler;


    package body Allocator is
        --
        -- There are essentially four choices here.
        -- 1. Call Apex_Ada_Krn_I.Alloc and .Free
        -- 2. Disable allocation by simply returning System.No_Addr
        -- 3. Use pragma Import to explicitly call __ADA_ALLOC
        --    (This is the same thing you get by calling Apex_Ada_Krn_I.Alloc,
        --    but without including Apex_Ada_Krn_I in the Ada closure of the
        --    harness.  This can be important for testing source
        --    runtime releases such as the MARK runtime.)
        -- 4. Write your own allocator (an exercise left to the user)
        --
        -- In any of these cases, Free can be safely no-op'd.  The harness is
        -- only used once in the life of the program, and you'll get your
        -- memory back when the program terminates.
        --
        function Aki_Alloc (Size_In_Bytes : Natural) return System.Address
           renames Apex_Ada_Krn_I.Alloc;
        procedure Aki_Free (Memory_Item : System.Address)
           renames Apex_Ada_Krn_I.Free;

        function Alloc (Size_In_Bytes : Natural) return System.Address is
        begin
            return Aki_Alloc (Size_In_Bytes);
        end Alloc;

        procedure Free (Memory_Item : in System.Address) is
        begin
            Aki_Free (Memory_Item);
        end Free;

    end Allocator;

    package body Mcdc_String_Utilities is

        function Hash_String (S : String) return Integer is
            A1, A2  : Natural := 0;
            Counter : Natural := 3;
        begin
            for J in S'Range loop
                A1      := A1 * (Character'Pos (Character'Last) + 1) +
                              Character'Pos (S (J));
                Counter := Counter - 1;

                if Counter = 0 then
                    A2 := (211 * A2 + A1) mod 10098131;
                    -- The significance of these constants is that
                    -- they are both prime, and
                    --
                    --    211 * 10098130 + 256**3;
                    --
                    -- is a little less than integer'last.

                    A1      := 0;
                    Counter := 3;
                end if;
            end loop;

            return 211 * A2 + A1;
        end Hash_String;


        function Locate (Fragment : Character; Within : String)
                        return Natural is
        begin
            for I in Within'Range loop
                if Fragment = Within (I) then
                    return I;
                end if;
            end loop;
            return 0;
        end Locate;


        function Reverse_Locate
                    (Fragment : Character; Within : String) return Natural is
        begin
            for I in reverse Within'Range loop
                if Fragment = Within (I) then
                    return I;
                end if;
            end loop;
            return 0;
        end Reverse_Locate;


        function Number_To_String (Value : Trace_Integer) return String is

            Sign       : Boolean          := False;
            Max_Length : constant Natural := (Trace_Integer'Size / 4) + 1;
            Result     : String (1 .. Max_Length);
            Next       : Natural          := Max_Length;

            procedure N2S (Num_Value : Trace_Integer) is
                Num : Trace_Integer := Num_Value;
            begin
                loop
                    if Num = 0 then
                        if Sign then
                            Result (Next) := '-';
                            Next          := Next - 1;
                        elsif Value = 0 then
                            Result (Next) := '0';
                            Next          := Next - 1;
                        end if;
                        exit;
                    else
                        Result (Next) := Base_Characters (Num rem 16);
                        Next          := Next - 1;
                        Num           := Num / 16;
                    end if;
                end loop;
            end N2S;

        begin
            if Value < 0 then
                Sign := True;
            end if;
            N2S (Value);
            return Result (Next + 1 .. Result'Last);
        end Number_To_String;


        function String_To_Number (Value : String) return Trace_Integer is
            Sign     : Trace_Integer := -1;
            Result   : Trace_Integer := 0;
            Ch       : Character;
            Char_Val : Trace_Integer;
            Prefix   : Boolean       := True;
        begin
            -- Since there is one more negative Integer value then positive
            -- the math in result := result*16 + val inside this loop is done
            -- calculating negative numbers, and then the result is multiplied by
            -- sign.  The value of Sign is the inverse of the actual sign of the
            -- number.
            for I in Value'Range loop
                Ch := Value (I);
                if Ch = ' ' then
                    if not Prefix then
                        exit;
                    end if;
                else
                    if Prefix and Ch = '-' then
                        Sign := +1;

                    elsif Ch /= '+' or else not Prefix then
                        case Ch is
                            when '0' .. '9' =>
                                Char_Val := Character'Pos (Ch) -
                                               Character'Pos ('0');
                            when 'A' .. 'F' =>
                                Char_Val := Character'Pos (Ch) -
                                               Character'Pos ('A') + 10;
                            when 'a' .. 'f' =>
                                Char_Val := Character'Pos (Ch) -
                                               Character'Pos ('a') + 10;
                            when others =>
                                raise Program_Error;
                        end case;
                        Result := Result * 16 - Char_Val;
                    end if;
                    Prefix := False;
                end if;
            end loop;

            return Result * Sign;
        end String_To_Number;


    end Mcdc_String_Utilities;


end Tm_Mcdc_Runtime_Support;
pragma Suppress_Elaboration_Checks;
