-- Copyright (C) 1995,2000,2002 Rational Software Corporation.
-- All rights reserved.

-- RESTRICTED RIGHTS LEGEND
--
-- Use, duplication, or disclosure by the Government and Boeing is subject to
-- restrictions as set forth in subparagraph (c)(1)(ii) of the Rights in
-- Technical Data and Computer Software clause of DFAR 252.227-7013 and in
-- similiar clauses in the FAR and NASA FAR Supplement.
------------------------------------------------------------------------------
-- PROPRIETARY AND CONFIDENTIAL INFORMATION OF RATIONAL SOFTWARE CORP.,
-- USE OR COPYING WITHOUT EXPRESS WRITTEN AUTHORIZATION IS STRICTLY PROHIBITED.
-- THIS MATERIAL IS PROTECTED AS AN UNPUBLISHED WORK UNDER THE U.S. COPYRIGHT
-- ACT OF 1976.  CREATED 1995.  ALL RIGHTS RESERVED.
------------------------------------------------------------------------------
--
-- File Name: tm_mcdc_score_support.2.ada
--

with Tm_Mcdc_Config;
with Tm_Mcdc_Allocation;
with Tm_Mcdc_Runtime_Support;
with Unchecked_Conversion;

package body Tm_Mcdc_Score_Support is

    pragma Suppress (Access_Check);
    pragma Suppress (Discriminant_Check);
    pragma Suppress (Index_Check);
    pragma Suppress (Length_Check);
    pragma Suppress (Range_Check);
    pragma Suppress (Storage_Check);

    -- Allocate a few extra bytes in the data pool to account for
    -- duplicate data elements.
    Decision_Array_Pool_Slack : constant := 100;
    Unit_Id_Node_Pool_Slack   : constant := 10;
    Unit_Node_Pool_Slack      : constant := 10;

    -- Allocate Decision_Array objects in chunks of 8.
    -- This works well for packed and unpacked arrays.
    -- Note that the Maximum_Number_Of_Conditions must account for the slack
    -- at the end of each 8-element chunk.
    -- The TM_instrumentation_stats script rounds up the size of each bitmap
    -- it finds to the next multiple of 8.
    -- If we ever want to allocate in chunks of other than 8, the script needs
    -- to be updated accordingly.
    Decision_Array_Element_Span : constant := 8;

    subtype Decision_Array_Element is Bool_Decision_Array
                                         (0 .. Decision_Array_Element_Span - 1);
    type Decision_Array_Handle is access Decision_Array_Element;
    for Decision_Array_Handle'Storage_Size use 0;

    Decision_Array_Pool_Length : constant :=
       (Tm_Mcdc_Config.Maximum_Number_Of_Conditions /
        Decision_Array_Element_Span) + Decision_Array_Pool_Slack;
    -- Give it a few extra bytes

    subtype Decision_Array_Pool_Index is
       Integer range 0 .. Decision_Array_Pool_Length;
    type Decision_Array_Pool_Storage is
       array (Decision_Array_Pool_Index) of Decision_Array_Element;

    package Decision_Array_Pool is
       new Tm_Mcdc_Allocation.Pool
              (Element      => Decision_Array_Element,
               Handle       => Decision_Array_Handle,
               Pool_Index   => Decision_Array_Pool_Index,
               Pool_Storage => Decision_Array_Pool_Storage);

    Decision_Array_Data : Decision_Array_Pool.Data :=
       (Storage => (others => (0 .. Decision_Array_Element_Span - 1 => False)),
        Locks   => (others => 0),
        Next    => 0);

    function Constructor (The_Length : Natural) return Unit_Decisions is
        function As_Unit_Decisions is
           new Unchecked_Conversion (Decision_Array_Handle, Unit_Decisions);

        Result : Unit_Decisions;
    begin
        if The_Length = 0 then
            return null;
        end if;

        declare
            Element_Count : Natural                      :=
               (The_Length + Decision_Array_Element_Span - 1) /
                  Decision_Array_Element_Span;
            Pool_Data_Ptr : Decision_Array_Pool.Data_Ptr :=
               Decision_Array_Pool.As_Data_Ptr (Decision_Array_Data'Address);
            The_Handle    : Decision_Array_Handle;
        begin
            The_Handle := Decision_Array_Pool.Allocate_Array
                             (Element_Count, Pool_Data_Ptr);
            Result     := As_Unit_Decisions (The_Handle);
            if Result /= null then
                Result (0 .. Condition_Id (The_Length - 1)) :=
                   (others => False);
            end if;
            return Result;
        end;
    end Constructor;

    procedure Destructor (The_Decisions : in out Unit_Decisions;
                          The_Length    : in out Natural) is
        function As_Decision_Array_Handle is
           new Unchecked_Conversion (Unit_Decisions, Decision_Array_Handle);
    begin
        if The_Length > 0 then
            declare
                Element_Count : Natural                      :=
                   (The_Length + Decision_Array_Element_Span - 1) /
                      Decision_Array_Element_Span;
                Pool_Data_Ptr : Decision_Array_Pool.Data_Ptr :=
                   Decision_Array_Pool.As_Data_Ptr
                      (Decision_Array_Data'Address);
                The_Handle    : Decision_Array_Handle        :=
                   As_Decision_Array_Handle (The_Decisions);
            begin
                Decision_Array_Pool.Deallocate_Array
                   (The_Handle, Element_Count, Pool_Data_Ptr);
                The_Decisions := null;
                The_Length    := 0;
            end;
        end if;
    end Destructor;

    package body Result_List is separate;
    package body Unit_Id_Map is separate;
    package body Unit_Map is separate;

    ---------------------------------------------------------------------------
    -- Interface routines
    ---------------------------------------------------------------------------

    procedure Destroy (The_Scorecard : in out Scorecard) is separate;
    -- Separate to avoid circular coding dependency.


    function Register (The_Score                 : Scorecard_Ref;
                       The_Unit                  : String;
                       The_Timestamp             : Integer;
                       The_Checksum              : Integer;
                       The_Decision_Array_Length : Natural;
                       The_Kind                  : Character)
                      return Unit_Info is
        The_Set : Unit_Map.Set;
    begin
        -- Check to see if unit already registered ...
        Unit_Map.Find (In_Map   => The_Score.The_Unit_Map,
                       The_Unit => The_Unit,
                       The_Set  => The_Set);

        -- If not found then insert it ...
        if Unit_Map.Is_Null (The_Set) then
            Unit_Map.Insert (Into_Map             => The_Score.The_Unit_Map,
                             The_Unit             => The_Unit,
                             The_Timestamp        => The_Timestamp,
                             The_Checksum         => The_Checksum,
                             The_Decisions_Length => The_Decision_Array_Length,
                             The_Decisions        =>
                                Constructor (The_Decision_Array_Length),
                             The_Kind             => The_Kind,
                             The_Set              => The_Set);

        elsif The_Set.The_Timestamp = 0 and then The_Timestamp /= 0 then
            -- If previous call (or this call) was a dummy registration call
            -- then update registration of the unit.  Dummy registration calls
            -- are made by default parameter initialization in compilation
            -- unit specifications.
            Unit_Map.Replace (The_Set              => The_Set,
                              The_Unit             => The_Unit,
                              The_Timestamp        => The_Timestamp,
                              The_Checksum         => The_Checksum,
                              The_Decisions_Length => The_Decision_Array_Length,
                              The_Decisions        =>
                                 Constructor (The_Decision_Array_Length),
                              The_Kind             => The_Kind);

        end if;

        return (The_Set.The_Id, The_Set.The_Decisions);
    end Register;


    procedure Register_Decision (The_Score   : Scorecard_Ref;
                                 The_Unit    : Unit_Id;
                                 The_Line    : Natural;
                                 The_Line_Id : Natural;
                                 The_Result  : Trace_Integer) is
        The_Result_List : Result_List.List;
        Found           : Boolean;
    begin
        Unit_Id_Map.Find (In_Map          => The_Score.The_Unit_Id_Map,
                          The_Unit        => The_Unit,
                          The_Line        => The_Line,
                          The_Line_Id     => The_Line_Id,
                          The_Result_List => The_Result_List,
                          Success         => Found);
        if Found then
            Result_List.Insert (On_List    => The_Result_List,
                                The_Result => The_Result);
        else
            Unit_Id_Map.Insert
               (Into_Map        => The_Score.The_Unit_Id_Map,
                The_Unit        => The_Unit,
                The_Line        => The_Line,
                The_Line_Id     => The_Line_Id,
                The_Result_List => Result_List.Make (The_Result));
        end if;
    end Register_Decision;


    function Is_Clean_Scorecard (The_Scorecard : Scorecard) return Boolean is
    begin
        --@@@ Logically want to check if there have been any hits against any
        --@@@ active units, but in reality if there are registered units then
        --@@@ we want to generate a scorecard result (even if null of hits).
        return Unit_Map.Is_Empty (The_Scorecard.The_Unit_Map);
    end Is_Clean_Scorecard;


    procedure Clean_Scorecard (The_Scorecard : in out Scorecard) is separate;
    -- Separate to avoid circular coding dependency.


    procedure Iterate (The_Score : Scorecard_Ref) is separate;
    -- Separate to avoid circular coding dependency.


    procedure Dump_Statistics is

        procedure Put (The_Stat : String; The_Value : Integer) is
        begin
            Process_Statistic ("* " & The_Stat & Integer'Image (The_Value));
        end Put;

    begin
        if Tm_Mcdc_Allocation.Total_Alloc_Size > 0 then
            Put ("Total Allocation:  ", Tm_Mcdc_Allocation.Total_Alloc_Size);
            Put ("Dealloc Allocation:", Tm_Mcdc_Allocation.Total_Dealloc_Size);
            Put ("Current Allocation:", Tm_Mcdc_Allocation.Current_Alloc_Size);
            Put ("Peak Allocation:   ", Tm_Mcdc_Allocation.Peak_Alloc_Size);
        end if;
    end Dump_Statistics;

    procedure Dump_Warnings is

        procedure Put (The_Msg : String) is
        begin
            Process_Warning (The_Msg);
        end Put;

    begin
        if Tm_Mcdc_Allocation.Unfulfilled_Allocs > 0 then
            Put ("Instrumentation ran out of memory, incomplete coverage " &
                 "for" & Integer'Image (Tm_Mcdc_Allocation.Unfulfilled_Allocs) &
                 " elements.");
            if Tm_Mcdc_Config.Allow_Dynamic_Allocation then
                Put ("(Dynamic allocation was enabled for instrumentation.)");
            else
                Put
                   ("You need to increase buffers in the Tm_Mcdc_Config package");
                Put ("or allow dynamic allocation for coverage.");
            end if;
        end if;
    end Dump_Warnings;

    function Convert (The_Value : Bool_Decision_Array) return String is
    begin
        --
        -- "Encode" in groups of four bits
        -- 1st bit is LSB of first char, 2 bit is next-lowest bit, etc.
        -- The output string is big-endian, but little-endian within each
        -- hex character.  Just a little bit confusing!
        --
        if The_Value'Length < 1 then
            return "0";
        end if;

        declare
            The_Length : Natural                               :=
               (The_Value'Length + 3) / 4;
            The_Result : String (1 .. The_Length);
            Di         : Natural                               := 1;
            Si         : Condition_Id                          :=
               The_Value'First;
            Sq         : constant Condition_Id                 :=
               The_Value'Last;
            The_Code   : constant array (0 .. 15) of Character :=
               "0123456789AbcDEF";
            D_Value    : Natural;
            pragma Suppress (Index_Check, On => The_Code);
            pragma Suppress (Index_Check, On => The_Result);
        begin
            The_Result (The_Result'Last) := 'Z';
            while Si <= Sq loop
                D_Value := 0;
                if The_Value (Si) then
                    D_Value := D_Value + 1;
                end if;
                Si := Si + 1;
                if Si <= Sq then
                    if The_Value (Si) then
                        D_Value := D_Value + 2;
                    end if;
                    Si := Si + 1;
                    if Si <= Sq then
                        if The_Value (Si) then
                            D_Value := D_Value + 4;
                        end if;
                        Si := Si + 1;
                        if Si <= Sq then
                            if The_Value (Si) then
                                D_Value := D_Value + 8;
                            end if;
                            Si := Si + 1;
                        end if;
                    end if;
                end if;
                The_Result (Di) := The_Code (D_Value);
                Di              := Di + 1;
            end loop;
            return The_Result;
        end;
    end Convert;


    function Convert (The_Value : String) return Bool_Decision_Array is

        type Character_Map is array (Character) of Natural;
        Get_Index : constant Character_Map := Character_Map'('0'       => 0,
                                                             '1'       => 1,
                                                             '2'       => 2,
                                                             '3'       => 3,
                                                             '4'       => 4,
                                                             '5'       => 5,
                                                             '6'       => 6,
                                                             '7'       => 7,
                                                             '8'       => 8,
                                                             '9'       => 9,
                                                             'A' | 'a' => 10,
                                                             'B' | 'b' => 11,
                                                             'C' | 'c' => 12,
                                                             'D' | 'd' => 13,
                                                             'E' | 'e' => 14,
                                                             'F' | 'f' => 15,
                                                             others    => 0);

    begin
        if The_Value = "" or else The_Value =
                                     String'(1 .. The_Value'Length => '0') then
            return (1 .. 0 => False);
        else
            declare
                The_Length : Condition_Id                           :=
                   The_Value'Length * 4;
                The_Result : Bool_Decision_Array (0 .. The_Length - 1);
                The_Next   : Condition_Id                           := 0;
                The_Index  : Natural                                :=
                   The_Value'First;
                subtype Short_Bool is Bool_Decision_Array (1 .. 4);
                Map        : constant array (0 .. 15) of Short_Bool :=
                   (0  => (False, False, False, False),
                    1  => (True, False, False, False),
                    2  => (False, True, False, False),
                    3  => (True, True, False, False),
                    4  => (False, False, True, False),
                    5  => (True, False, True, False),
                    6  => (False, True, True, False),
                    7  => (True, True, True, False),
                    8  => (False, False, False, True),
                    9  => (True, False, False, True),
                    10 => (False, True, False, True),
                    11 => (True, True, False, True),
                    12 => (False, False, True, True),
                    13 => (True, False, True, True),
                    14 => (False, True, True, True),
                    15 => (True, True, True, True));
            begin
                while The_Index <= The_Value'Last loop
                    The_Result (The_Next .. The_Next + 3) :=
                       Map (Get_Index (The_Value (The_Index)));
                    The_Next                              := The_Next + 4;
                    The_Index                             := The_Index + 1;
                end loop;
                return The_Result;
            end;
        end if;
    end Convert;


end Tm_Mcdc_Score_Support;
