--+--------------------------------------------------------------------------+
--| IBM Confidential                                                         |
--| OCO Source Materials                                                     |
--| Rational Apex                                                            |
--| (C) Copyright IBM Corp. 2008 All Rights Reserved                         |
--|                                                                          |
--| The source code for this program is not published or otherwise divested  |
--| of its trade secrets, irrespective of what has been deposited with the   |
--| U.S. Copyright Office.                                                   |
--+--------------------------------------------------------------------------+
--| Copyright 1992 - 2002 Rational Software Corporation. All Rights Reserved.|
--| This software contains proprietary and confidential information of       |
--| Rational and its suppliers. Use, disclosure or reproduction is           |
--| prohibited without the prior express written consent of Rational.        |
--+--------------------------------------------------------------------------+
--
with Tms_Arguments;
with Apex_Unix_Types;

package Tms_Communications is

    --
    -- This package is used as the conduit for communications between the
    -- test scripts (or separately executing jobs) and the execution server
    -- which processes the requests.  For requests that require a response
    -- the execution server will fill in a response and respond to the request.
    -- In reality, all requests are responded to so that the program stays
    -- synchronized with the associated server program.
    --
    -- Implementation of the communications is performed through a socket
    -- allocated by the server (before execution of the script is started)
    -- The hostid:port_number pair of this socket is passed to the
    -- script via an environment variable (the variable is defined in the
    -- current process, and it is assumed that it is passed on to any
    -- child processes forked from this parent).  The
    -- script connects to the socket to complete the connection.  A test 
    -- using TMS without the execution server running will encounter a void
    -- environment variable setting and simply ignore requests.
    --
    -- The original version of this package contained a request record that
    -- was passed through the socket (in place of the Byte_Array used now).
    -- This implementation proved expensive in terms of the sizes of requests
    -- being passed.  Attempts to reduce the size of requests increased the
    -- number of discriminants associated with the request, which affected
    -- all requests instead of just those requiring the discriminants.  The
    -- lesser of the evils was deemed a separate encoding package that created
    -- simple byte arrays that are passed.  The down side is ofcourse the 
    -- immediate argument checking that was handled by the request record
    -- now being done on both sides of the socket.  On the up side, this 
    -- implementation is much more flexible and can easily be used for other
    -- uses in the future.
    --

    subtype Byte_Array is Tms_Arguments.Byte_Array;

    --
    -- Name of the environment variable to pass from parent process
    -- to child process containing the port id for communications
    --
    Port_Id_Env : constant String := "TESTMATE_COMM_PORT";

    --*******************************
    --**
    --** The following should be considered a customization.  On Hp,
    --** reading from a socket that is still open may return immediately,
    --** so in this case we will loop until the number of requested bytes
    --** has been received (which may cause looping forever in some 
    --** erroneous cases).  Currently, an environment variable APEX_ARCH
    --** is used to detect execution on hppa_hpux systems and to set it
    --** as false.  See the elaboration for this package for actual determination
    --** of this value.
    --**

    Treat_0_As_Eof : Boolean := True;

    --**
    --*******************************


    package Communicate is

        End_Of_File : exception;
        --
        -- When an end of file mark is received the following exception is 
        -- raised to alert the caller that the connection has been severed.
        --

        procedure Send (Request : Byte_Array);
        --
        -- Writes a request to the active socket.  On the server side this
        -- is actually a response to a client request.  On the client side
        -- this is simply a request.
        -- 

        function Receive return Byte_Array;
        --
        -- Simple receive of a call_request.  On the server side this is
        -- actually the initiation of a request to be performed.  On the
        -- client side this is the receiving of a reply to a already 
        -- performed request.
        --

        generic
            with procedure Process (Request : Byte_Array);
        procedure Receive_Request;
        --
        -- This receive is just like the receive above with the distinction
        -- that the receive is passed a routine that will handle the request.
        -- This routine will be used by the server to handle client requests
        -- instead of passing this result on itself.  (The name has been 
        -- changed from RECEIVE because the compiler considered it a 
        -- "co-resident homograph" of the receive above; which ofcourse is
        -- total mallarky).
        --

    end Communicate;



    generic
        with procedure Export_Port_Info (Server : Apex_Unix_Types.Int;
                                         Port   : Apex_Unix_Types.Int);
    package Server is
        --
        -- Support for iterative or concurrent servers.  A control socket
        -- is assigned to an ephemeral port, and that port is passed to
        -- the child (via a mechanism defined by Export_Port_Info and 
        -- Import_Port_Info).  The child (client) process requests a connection to the
        -- server at this port.  The server creates a new socket for the 
        -- connection.  Iterative servers will perform necessary communications
        -- on this socket, then close it and return to listening to connections
        -- on the control socket.  Concurrent servers will instead fork off
        -- a child process to handle the new connection, and immediately
        -- listen for new connection requests on the control socket.
        --
        -- TestMate currently employs an iterative server.
        --


        procedure Setup (Control_Socket : out Apex_Unix_Types.Int);
        --
        -- Allocates a control socket to be used in communications with the 
        -- clients of this server, and binds an ephemeral port id to it.  
        -- Export_Port_Info is called to register the hostname/port combination
        -- for the client to access.  The socket's fd is returned to the 
        -- server.
        --


        function Connect return Apex_Unix_Types.Int;
        --
        -- The server calls this routine to accept a connection
        -- with a client, on the socket returned by Control_Socket.  A new 
        -- socket is created for the connection, and the new socket's fd 
        -- is returned.  This allows concurrent server to be created,
        -- where a new process is forked to handle the connection while
        -- the original process listens for new connections on the 
        -- control socket.
        --
        -- Control_Socket must have been called prior to this call.
        --


        procedure Disconnect;
        --
        -- The server calls this routine to close a connection with
        -- a client.  The Control_Socket remains unaffected.
        --

    end Server;




    generic
        with procedure Import_Port_Info (Server : out Apex_Unix_Types.Int;
                                         Port   : out Apex_Unix_Types.Int);
    package Client is

        procedure Connect;
        --
        -- Retrieves the server's port id via Import_Port_Info and 
        -- attempts to connect to the server.  If the environment
        -- variable is not set or contains an invalid value then all
        -- requests made by the client are ignored.
        --

        procedure Send (Request : Byte_Array);
        --
        -- Same as Communicate.Send, except first connects to the server if
        -- needed.  
        --

        procedure Disconnect;
        --
        -- Resets state information.  Actually never called and not necessary
        -- to call because the state information simply goes away when the
        -- process terminates anyway.  Here for consistency anyway.
        --

        function Connected return Boolean;
        --
        -- Returns true if the client is connected to the server.  If
        -- false, then either the connection could not be made or has
        -- dissappeared, or the client was not started by the server.
        --

        function Bastard return Boolean;
        --
        -- Returns true if the client was not started by the server.
        -- ( Bastard implies not Connected)
        --

    end Client;


    procedure Trace (Msg : String);
    -- debugging


    procedure Report_Error (Msg : String; Force_Display : Boolean := False);
    --
    -- Write a message to stderr, prepended with "!!!"
    -- By default, message is suppressed for serverless clients unless
    -- Force_Display is set.  This allows test drivers using package Tms
    -- to be run without extraneous messages to standard error.
    --

    procedure Force_Serverless_Connectcheck (Force : Boolean := False);
    pragma Inline (Force_Serverless_Connectcheck);
    --
    -- For a serverless client, show an initial error message indicating that
    -- there is no connection.  Normally, this check is suppressed.
    --

end Tms_Communications;
--[api] pragma Api(testmate);
