--+--------------------------------------------------------------------------+
--| IBM Confidential                                                         |
--| OCO Source Materials                                                     |
--| Rational Apex                                                            |
--| (C) Copyright IBM Corp. 2008 All Rights Reserved                         |
--|                                                                          |
--| The source code for this program is not published or otherwise divested  |
--| of its trade secrets, irrespective of what has been deposited with the   |
--| U.S. Copyright Office.                                                   |
--+--------------------------------------------------------------------------+
--| Copyright 1992 - 2002 Rational Software Corporation. All Rights Reserved.|
--| This software contains proprietary and confidential information of       |
--| Rational and its suppliers. Use, disclosure or reproduction is           |
--| prohibited without the prior express written consent of Rational.        |
--+--------------------------------------------------------------------------+
--
with System;
with Tm_Simple_Io;
with Unchecked_Conversion;
with Apex_Unix_Operations;
with Unix_Operations_Extension;
with Apex_Unix_Errno;
with Apex_Unix_Constants_Nonstandard;
with Apex_Unix_Structures_Nonstandard;
with Apex_Unix_Network_Operations;
with Apex_Unix_Types;

package body Tms_Communications is

    package Ta renames Tms_Arguments;

    function "&" (L, R : Ta.Byte_Array) return Ta.Byte_Array renames Ta."&";
    function "=" (L, R : Ta.Routine_Names) return Boolean renames Ta."=";
    function "=" (L, R : Apex_Unix_Types.Char_Ptr) return Boolean
       renames Apex_Unix_Types."=";
    function "=" (L, R : Apex_Unix_Types.Ushort) return Boolean
       renames Apex_Unix_Types."=";

    function "=" (Left, Right : Apex_Unix_Types.Int) return Boolean
       renames Apex_Unix_Types."=";
    function ">=" (Left, Right : Apex_Unix_Types.Int) return Boolean
       renames Apex_Unix_Types.">=";
    function "<" (Left, Right : Apex_Unix_Types.Int) return Boolean
       renames Apex_Unix_Types."<";
    function "-" (Left, Right : Apex_Unix_Types.Int) return Apex_Unix_Types.Int
       renames Apex_Unix_Types."-";
    function "+" (Left, Right : Apex_Unix_Types.Int) return Apex_Unix_Types.Int
       renames Apex_Unix_Types."+";
    function "/" (Left, Right : Apex_Unix_Types.Int) return Apex_Unix_Types.Int
       renames Apex_Unix_Types."/";
    function "-" (Unary : Apex_Unix_Types.Int) return Apex_Unix_Types.Int
       renames Apex_Unix_Types."-";

    function To_Inaddr is
       new Unchecked_Conversion
              (Apex_Unix_Types.Int, Apex_Unix_Structures_Nonstandard.In_Addr);

    function Decode_String
                (From_Call : Ta.Byte_Array; Parameter : Natural) return String
       renames Ta.Decode_String;


    subtype Size_Array is Ta.Byte_Array (1 .. 2);

    type Users is (Undefined, Tms_Client, Tms_Server, Tms_Serverless_Client);

    User : Users := Undefined;

    Control_Sockfd : Apex_Unix_Types.Int := 0;
    Data_Sockfd    : Apex_Unix_Types.Int := 0;

    --
    -- The Control_Sockfd is the 'well known' server socket.  When 
    -- the client connects to this socket, the accept() call in the server 
    -- will return a new socket, Data_Sockfd, with which to 
    -- communicate to the client.  On the client side, Data_Sockfd is
    -- the socket used to communicate with the server.
    --


    Control_Sockaddr : Apex_Unix_Structures_Nonstandard.Sockaddr_In;
    Data_Sockaddr    : Apex_Unix_Structures_Nonstandard.Sockaddr_In;
    Trace_Enabled    : Boolean := False;

    Link_Down_Reported             : Boolean := False;
    Enable_Serverless_Connectcheck : Boolean := False;

    Sockaddr_Len : constant Apex_Unix_Types.Int :=
       Control_Sockaddr'Size / System.Storage_Unit;

    -------------------  Support routines   -------------------


    function "+" (L : Apex_Unix_Types.Char_Ptr; R : Apex_Unix_Types.Int)
                 return Apex_Unix_Types.Char_Ptr is
        function To_Address is
           new Unchecked_Conversion (Apex_Unix_Types.Char_Ptr, System.Address);
    begin
        return Apex_Unix_Types.To_Char_Ptr
                  (System."+" (To_Address (L), Integer (R)));
    end "+";

    function To_Size_Array (Value : Natural) return Size_Array is
        Result : Size_Array := (Apex_Unix_Types.Char (Value / 128),
                                Apex_Unix_Types.Char (Value rem 128));
    begin
        return Result;
    end To_Size_Array;

    function From_Size_Array (Arr : Size_Array) return Natural is
        Result : Natural := (Natural (Arr (1)) * 128) + Natural (Arr (2));
    begin
        return Result;
    end From_Size_Array;


    function Get_Host_Name return String is
        Max_Host_Name_Len : constant := 32;
        -- MAXHOSTNAMELEN from <sys/param.h>

        Host_Name : String (1 .. Max_Host_Name_Len + 1);
    begin
        Unix_Operations_Extension.Gethostname
           (Apex_Unix_Types.To_Char_Ptr (Host_Name (1)'Address),
            Max_Host_Name_Len);
        return Apex_Unix_Types.To_String
                  (Apex_Unix_Types.To_Char_Ptr
                      (Host_Name (Host_Name'First)'Address));
    end Get_Host_Name;


    function Host_Internet_Address return Apex_Unix_Types.Int is
        C_Host_Name      : constant String                              :=
           Get_Host_Name & Ascii.Nul;
        Internet_Hostent : Apex_Unix_Structures_Nonstandard.Hostent_Ptr :=
           Apex_Unix_Network_Operations.Gethostbyname
              (Apex_Unix_Types.To_Char_Ptr
                  (C_Host_Name (C_Host_Name'First)'Address));
        Ip               : Apex_Unix_Types.Int;
    begin
        if Apex_Unix_Structures_Nonstandard."=" (Internet_Hostent, null) then
            Report_Error ("Could not find address for " & C_Host_Name);
        end if;
        Ip := Apex_Unix_Types.To_Int_Ptr
                 (Internet_Hostent.H_Addr_List.all.all'Address).all;
        return Ip;
    end Host_Internet_Address;


    -------------------  Interface routines -------------------


    package body Communicate is

        procedure Send (Request : Byte_Array) is
            Msg : constant Byte_Array :=
               To_Size_Array (Request'Length) & Request;
        begin
            if Unix_Operations_Extension.Send
                  (S     => Data_Sockfd,
                   Msg   => Apex_Unix_Types.To_Char_Ptr
                               (Msg (Msg'First)'Address),
                   Len   => Msg'Length,
                   Flags => 0) = -1 then
                Report_Error ("Send " & Ta.Routine_Names'Image
                                           (Ta.Routine (Request)));
                if Ta.Routine (Request) = Ta.Tms_Put_Line then
                    -- at least allow TMS.Put_Line to work in a rudimentary way...
                    Tm_Simple_Io.Put_Line (Decode_String (Request, 1));
                end if;
            end if;
        exception
            when others =>
                Report_Error ("Send Error");
        end Send;

        procedure Recv (Buf      : Apex_Unix_Types.Char_Ptr;
                        Len      : Natural;
                        On_Error : String) is
            Buffer       : Apex_Unix_Types.Char_Ptr := Buf;
            Amount_Recvd : Apex_Unix_Types.Int      := 0;
            Total_Recvd  : Apex_Unix_Types.Int      := 0;
        begin
            while Total_Recvd /= Apex_Unix_Types.Int (Len) loop
                Buffer       := Buffer + Amount_Recvd;
                Amount_Recvd :=
                   Unix_Operations_Extension.Recv
                      (S     => Data_Sockfd,
                       Buf   => Buffer,
                       Len   => Apex_Unix_Types.Int (Len) - Total_Recvd,
                       Flags => 0);
                if Amount_Recvd = -1 then
                    if Apex_Unix_Operations.Get_Errno /=
                       Apex_Unix_Errno.Eintr then
                        Report_Error ("Receive Error on " & On_Error);
                        exit;
                    else
                        -- Interrupted while waiting for read, nothing read yet!
                        Amount_Recvd := 0;
                    end if;
                elsif Amount_Recvd = 0 then
                    -- End of file encountered ...
                    if Treat_0_As_Eof then
                        -- Normal UNIX (except HP when not doing Select ...)
                        raise End_Of_File;
                    else
                        Trace ("Recv: 0 Amount_Recvd - not as EOF");
                    end if;
                end if;
                Total_Recvd := Total_Recvd + Amount_Recvd;
            end loop;
        end Recv;

        function Receive_Size return Natural is
            Sz : Size_Array := (0, 0);
        begin
            Recv (Buf      => Apex_Unix_Types.To_Char_Ptr (Sz (1)'Address),
                  Len      => 2,
                  On_Error => "Receive Size");
            return From_Size_Array (Sz);
        exception
            when End_Of_File =>
                raise;
            when others =>
                Report_Error
                   ("Buffer Size to Conversion error - Request dropped");
                return 0;
        end Receive_Size;

        function Receive return Byte_Array is
            Buf : Byte_Array (1 .. Receive_Size);
        begin
            if Buf'Length /= 0 then
                Recv (Buf      => Apex_Unix_Types.To_Char_Ptr (Buf (1)'Address),
                      Len      => Buf'Length,
                      On_Error => "User Receive");
            end if;
            return Buf;
        exception
            when End_Of_File =>
                raise;
            when others =>
                Report_Error ("Buffer to Call Request error");
                raise;
        end Receive;

        procedure Receive_Request is
            Buf : Byte_Array (1 .. Receive_Size);
        begin
            if Buf'Length /= 0 then
                Recv (Buf      => Apex_Unix_Types.To_Char_Ptr (Buf (1)'Address),
                      Len      => Buf'Length,
                      On_Error => "User Receive");
                Process (Buf);
            end if;
        exception
            when End_Of_File =>
                raise;
            when others =>
                Report_Error ("Buffer to Call Request error");
        end Receive_Request;

    end Communicate;


    ------------------------------------------------------------------------


    package body Server is

        Port_Id : Apex_Unix_Types.Int := 0;

        procedure Setup (Control_Socket : out Apex_Unix_Types.Int) is
            Info_Len : Apex_Unix_Types.Int;
        begin
            -- Get a socket and bind it to an address; then create an 
            -- environment variable with the address as the value.
            Control_Sockfd := Apex_Unix_Network_Operations.Socket
                                 (Domain   => Apex_Unix_Constants_Nonstandard.
                                                 Unix_Socket.Af_Inet,
                                  Typ      => Apex_Unix_Constants_Nonstandard.
                                                 Unix_Socket.Sock_Stream,
                                  Protocol => 0);
            if Control_Sockfd = -1 then
                Report_Error ("Sever socket allocation failed");
            end if;

            Control_Sockaddr          :=
               Unix_Operations_Extension.Default_Sockaddr_In;
            Control_Sockaddr.Sin_Addr := To_Inaddr (Host_Internet_Address);

            if Apex_Unix_Network_Operations.Bind
                  (S       => Control_Sockfd,
                   Name    => Apex_Unix_Structures_Nonstandard.To_Sockaddr_Ptr
                                 (Control_Sockaddr'Address),
                   Namelen => Sockaddr_Len) = -1 then
                Report_Error ("Server bind failed");
            end if;

            if Control_Sockaddr.Sin_Port = 0 then
                -- Use getsockname to find out what port we were given
                Info_Len := Control_Sockaddr'Size / System.Storage_Unit;
                if Unix_Operations_Extension.Getsockname
                      (S       => Control_Sockfd,
                       Addr    =>
                          Apex_Unix_Structures_Nonstandard.To_Sockaddr_In_Ptr
                             (Control_Sockaddr'Address),
                       Addrlen =>
                          Apex_Unix_Types.To_Int_Ptr (Info_Len'Address)) =
                   -1 then
                    Report_Error ("Server getsockname failed");
                end if;
            end if;

            if Apex_Unix_Network_Operations.Listen
                  (S => Control_Sockfd, Backlog => 5) = -1 then
                Report_Error ("Server socket listen failure");
            end if;

            Control_Socket := Control_Sockfd;
            Port_Id        := Apex_Unix_Types.Int (Control_Sockaddr.Sin_Port);

            Export_Port_Info (Host_Internet_Address, Port_Id);
        end Setup;


        function Connect return Apex_Unix_Types.Int is
            Sa_Length : Apex_Unix_Types.Int :=
               Control_Sockaddr'Size / System.Storage_Unit;
            Errno     : Apex_Unix_Types.Int;
        begin
            -- Make the control socket non-blocking, don't wait if there is 
            -- nothing to read from the socket.  Note, we should only attempt 
            -- to read from the server side of the connection when something 
            -- is waiting to be read, this is insured through the use of the
            -- SELECT system call in the server.
            -- Fio_Arg := 1;
            -- if Fio.Ioctl (Fildes => Control_Sockfd,
            --               Request => Fio.Fionbio, 
            --               Arg => unix_types.To_Int_Ptr (Fio_Arg'Address)) = -1 then
            --     Report_Error ("Server could not connect to target program");
            -- end if;


            for Tries in 1 .. 60 loop
                Data_Sockfd :=
                   Apex_Unix_Network_Operations.Accpt
                      (S       => Control_Sockfd,
                       Addr    => Apex_Unix_Structures_Nonstandard.
                                     To_Sockaddr_Ptr (Control_Sockaddr'Address),
                       Addrlen => Apex_Unix_Types.To_Int_Ptr
                                     (Sa_Length'Address));
                exit when Data_Sockfd >= 0;

                -- exit when error is not interrupted
                Errno := Apex_Unix_Operations.Get_Errno;
                exit when Errno /= Apex_Unix_Errno.Eintr and
                             Errno /= Apex_Unix_Errno.Ewouldblock;
                delay 0.5;
            end loop;
            if Data_Sockfd < 0 then
                Report_Error ("Server socket accept failure");
            end if;
            Trace ("TMS_Communications.Server.Connect (sockfd is " &
                   Apex_Unix_Types.Int'Image (Data_Sockfd) & ")");
            return Data_Sockfd;
        end Connect;


        procedure Disconnect is
        begin
            if Apex_Unix_Operations.Close (Data_Sockfd) = -1 then
                Report_Error ("Closing server socket");
            end if;
            Data_Sockfd := 0;
        end Disconnect;

    begin
        User := Tms_Server;
    end Server;


    ------------------------------------------------------------------------


    package body Client is

        Client_Connected : Boolean := False;
        No_Server        : Boolean := True;

        procedure Connect is
            Server_Addr : Apex_Unix_Structures_Nonstandard.Sockaddr_In;
            Info_Length : Apex_Unix_Types.Int;
            Server      : Apex_Unix_Types.Int;
            Port        : Apex_Unix_Types.Int;
        begin
            Import_Port_Info (Server, Port);
            No_Server := (Server = 0 and Port = 0);

            if No_Server then
                User             := Tms_Serverless_Client;
                Client_Connected := False;
                Data_Sockfd      := -1;
                Trace ("We are either the server or a serverless tms client");
            else
                Trace ("Client connecting to server " &
                       Apex_Unix_Types.Int'Image (Server) & ", port " &
                       Apex_Unix_Types.Int'Image (Port));

                -- Allocate a socket id, then connect it to the server through
                -- the environment variable passed to this shell by the server
                -- (parent).
                Data_Sockfd := Apex_Unix_Network_Operations.Socket
                                  (Domain   => Apex_Unix_Constants_Nonstandard.
                                                  Unix_Socket.Af_Inet,
                                   Typ      => Apex_Unix_Constants_Nonstandard.
                                                  Unix_Socket.Sock_Stream,
                                   Protocol => 0);
                if Data_Sockfd = -1 then
                    Report_Error ("Client socket allocation failed");
                end if;

                Server_Addr          :=
                   Unix_Operations_Extension.Default_Sockaddr_In;
                Server_Addr.Sin_Port := Apex_Unix_Types.Ushort (Port);
                Server_Addr.Sin_Addr := To_Inaddr (Server);

                if Apex_Unix_Network_Operations.Connect
                      (S       => Data_Sockfd,
                       Name    =>
                          Apex_Unix_Structures_Nonstandard.To_Sockaddr_Ptr
                             (Server_Addr'Address),
                       Namelen => Server_Addr'Size / System.Storage_Unit) /=
                   0 then
                    -- Report_Error ("Client connect to server failure");
                    Client_Connected := False;
                else
                    -- Use getsockname to find out what port we were given
                    Info_Length := Data_Sockaddr'Size / System.Storage_Unit;
                    if Unix_Operations_Extension.Getsockname
                          (S       => Data_Sockfd,
                           Addr    => Apex_Unix_Structures_Nonstandard.
                                         To_Sockaddr_In_Ptr
                                         (Data_Sockaddr'Address),
                           Addrlen => Apex_Unix_Types.To_Int_Ptr
                                         (Info_Length'Address)) = -1 then
                        Report_Error ("Client getsockname failed");
                        Client_Connected := False;
                    else
                        Client_Connected := True;
                    end if;
                end if;

                Trace ("Client connected to server, client port =>" &
                       Apex_Unix_Types.Ushort'Image (Data_Sockaddr.Sin_Port) &
                       ";  Addr =>" &
                       Apex_Unix_Structures_Nonstandard.In_Addr'Image
                          (Data_Sockaddr.Sin_Addr) &
                       "; Sockfd => " &
                       Apex_Unix_Types.Int'Image (Data_Sockfd));
            end if;
        end Connect;


        procedure Disconnect is
        begin
            if Apex_Unix_Operations.Close (Data_Sockfd) = -1 then
                Report_Error ("Closing client socket");
            end if;
            Client_Connected := False;
        end Disconnect;


        procedure Send (Request : Byte_Array) is
        begin
            if not Client_Connected then
                Connect;
            end if;
            Communicate.Send (Request);
        end Send;


        function Connected return Boolean is
        begin
            return Client_Connected and not Link_Down_Reported;
        end Connected;


        function Bastard return Boolean is
        begin
            return No_Server;
        end Bastard;

    begin
        User := Tms_Client;
        --
        -- Do NOT connect the client at elaboration time, since this precludes
        -- the use of "tms" from within "testmate_shell" scripts run as test
        -- drivers (among other things).  Problem is that you cannot have two
        -- connected clients concurrently, and if the clients connect at
        -- elaboration time, then of course this is violated
        --
        -- Instead, the users should call Client.Send
        --
    end Client;


    procedure Report_Error (Msg : String; Force_Display : Boolean := False) is
        procedure Put is
        begin
            if Apex_Unix_Operations.Get_Errno = 0 then
                Tm_Simple_Io.Put_Line ("!!! " & Msg, Tm_Simple_Io.Stderr);
            else
                Tm_Simple_Io.Put_Line
                   ("!!! " & Msg & " (Unix errno =" &
                    Apex_Unix_Types.Int'Image (Apex_Unix_Operations.Get_Errno) &
                    ')', Tm_Simple_Io.Stderr);
            end if;
        end Put;
    begin
        -- Report link down if that is the case
        if User /= Tms_Serverless_Client or Enable_Serverless_Connectcheck then
            if Apex_Unix_Operations.Get_Errno = Apex_Unix_Errno.Enotconn or else
               Apex_Unix_Operations.Get_Errno = Apex_Unix_Errno.Enotsock then
                if not Link_Down_Reported then
                    if User = Tms_Server then
                        Tm_Simple_Io.Put_Line
                           ("!!! TestMate server: no connection to Tms client",
                            Tm_Simple_Io.Stderr);
                    elsif User = Tms_Client then
                        --or User = Tms_Serverless_Client then
                        Tm_Simple_Io.Put_Line
                           ("!!! Tms client: no connection to TestMate",
                            Tm_Simple_Io.Stderr);
                        if User = Tms_Client then
                            Tm_Simple_Io.Put_Line
                               ("... (is environment variable TESTMATE_COMM_PORT corrupted?)",
                                Tm_Simple_Io.Stderr);
                        end if;
                    elsif User = Tms_Serverless_Client then
                        Tm_Simple_Io.Put_Line
                           ("!!! Tms client: no connection to TestMate" &
                            " (must execute in test context)",
                            Tm_Simple_Io.Stderr);
                    end if;
                    if User /= Tms_Serverless_Client then
                        Tm_Simple_Io.Put_Line
                           ("!!! Continuing with limited functionality",
                            Tm_Simple_Io.Stderr);
                    end if;
                    Link_Down_Reported := True;
                end if;
            end if;
        end if;

        -- For serverless client, link down is generally enough
        if User /= Tms_Serverless_Client or Force_Display then
            Put;
        end if;
    end Report_Error;

    procedure Force_Serverless_Connectcheck (Force : Boolean := False) is
    begin
        Enable_Serverless_Connectcheck := Force;
    end Force_Serverless_Connectcheck;

    procedure Trace (Msg : String) is
    begin
        if Trace_Enabled then
            Tm_Simple_Io.Put_Line ("??? " & Msg, Tm_Simple_Io.Stderr);
        end if;
    end Trace;


begin
    declare
        C_Env      : constant String          :=
           "APEX_TM_LOG_INTERNAL_TRACE" & Ascii.Nul;
        Trace_Env  : Apex_Unix_Types.Char_Ptr :=
           Unix_Operations_Extension.Getenv
              (Apex_Unix_Types.To_Char_Ptr (C_Env (C_Env'First)'Address));
        C_Arch_Env : constant String          := "APEX_ARCH" & Ascii.Nul;
        Apex_Arch  : Apex_Unix_Types.Char_Ptr :=
           Unix_Operations_Extension.Getenv
              (Apex_Unix_Types.To_Char_Ptr
                  (C_Arch_Env (C_Arch_Env'First)'Address));
    begin
        Trace_Enabled := Trace_Env /= null;

        -- On hppa_hpux, sockets can return with 0 bytes, even though there
        -- is still more to read...detect this platform and act accordingly
        Treat_0_As_Eof :=
           Apex_Arch = null or else
              (Apex_Unix_Types.To_String (Apex_Arch) /= "hppa_hpux");
    end;

end Tms_Communications;
pragma Suppress_Elaboration_Checks;
