-- Copyright (C) 1993-2002 Rational Software Corporation.  All rights reserved.
--
-- The use of this system is subject to the software license terms and
-- conditions agreed upon between Rational and the Customer.
--
--
--                     -- RESTRICTED RIGHTS LEGEND --
--
-- Use, duplication, or disclosure by the U.S. Government is subject to
-- restrictions as set forth in in the applicable license agreement and
-- as provided in DFARS 277.7202-1(a) and 277.7202-3(a)(1995), DFARS
-- 252.227-7013(c)(1)(ii) (Oct 1988), FAR 12.212(a)(1995), FAR 52.227-19,
-- or FAR 227-14, as applicable.
--
-- Rational Software Corporation
-- 18880 Homestead Road
-- Cupertino, California, USA 95014
-- http://www.rational.com   mailto:support@rational.com
--
-- This source code is proprietary and confidential information of
-- Rational Software Corporation.  Use or copying without express written
-- authorization is strictly prohibited.  This material is protected under
-- the U.S. Copyright Act of 1976 and similar acts in other countries.
-- Created 1993.  All rights reserved.
--
-----------------------------------------------------------------------------

with System;
with Calendar;
package Tms is

    ------------------------------------------------------------------------
    --
    -- The TMS package is the primary API for dynamically reporting
    -- test results to TestMate's servers, entering messages in the
    -- test log, registering assertions, etc.  Several queries are
    -- also provided to allow test drivers to obtain information on
    -- the state of the currently executing test list or test case.
    --
    -- In general, calls to these routines while not executing under
    -- control of TestMate are ignored.
    --


    ------------------------------------------------------------------------
    --
    -- TEST RESULTS
    --
    --
    -- Test results are defined as strings, allowing for user-defined
    -- results (see User's Guide for information on defining new
    -- execution result values).  Functions are provided for the
    -- predefined execution results.
    --
    subtype Test_Result is String;


    --
    -- The following are the result kinds predefined by TestMate.
    --
    function Pass return Test_Result;
    function Fail return Test_Result;
    function Unknown return Test_Result;

    function Failed_Assertion return Test_Result;
    function Failed_Precondition return Test_Result;
    function Fail_With_Exception return Test_Result;
    function Terminated return Test_Result;
    function Timeout_Cpu return Test_Result;
    function Timeout_Elapsed return Test_Result;
    function Unhandled_Signal return Test_Result;
    function Not_Applicable return Test_Result;


    --
    -- Test Result Registration (Direct)
    --
    -- There are two basic techniques for registering test results: Setting
    -- the result directly, or inferring the result from assertions.
    -- Setting the result directly is easier, but less flexible.
    --
    --
    procedure Register_Result (The_Result        : Tms.Test_Result;
                               For_The_Test_Case : String := "");
    --
    -- Registers the result for the indicated test case.  A value of
    -- "" for test case name means the current test case.  Register_Result
    -- will clear the value of any previous Register_Result call (and any
    -- previous assertions, too).
    --
    -- Usage example:
    --
    --    if Device.Power_Up_Test_Ok then
    --         Tms.Register_Result(Tms.Pass);
    --    else
    --         Tms.Register_Result(Tms.Fail);
    --    end if;
    --
    -- Usage example of registering a user-defined test result:
    --
    --    Tms.Register_Result( Tms.Test_Result("BIT_PASSED"));
    --
    -- See the User's manual for more information regarding user-defined
    -- test results.
    --

    --
    -- Test Result Registration (Using Assertions)
    --
    -- Assertions provide an alternative mechanism for registering test
    -- results.  Rather than explicitly calling Register_Result, a test
    -- may make repeated assertions (boolean conditions that should hold
    -- true) during a test.  If any of these assertions fail (i.e. are
    -- evaluated as false) in a particular test case, then TestMate will
    -- automatically register a Failed_Assertion result.
    --
    -- Failed assertions do NOT terminate a test prematurely.
    --

    type Assertion_Status is (None, All_True, Some_False);

    type Log_Mode is (Log_All, Failures_Only, None);

    Default_Log_Mode : Log_Mode := Log_All;

    procedure Assert (Condition     : Boolean;
                      Description   : String   := "";
                      Log_Entry_For : Log_Mode := Default_Log_Mode);
    --
    -- If Assert is called at least once within the driver or evaluation
    -- actions, the test result is inferred from the assertions:
    --
    --     PASS              All assertions were True
    --     FAILED_ASSERTION  At least one assertion was false
    --     UNKNOWN           No assertions were made
    --
    -- By default, TestMate will log all assertions.  This behavior can
    -- be changed by changing the Default_Log_Mode or specifying the
    -- value of Log_Entry_For.
    --
    --    Log_Failures_Only causes only failed assertions to be logged
    --    Log_None causes no assertions to be logged
    --    Log_All causes all assertions to be logged
    --
    -- Usage example:
    --
    --    Tms.Assert( Device.Stat = 0, "Ensure device stat is 0");
    --
    -- or:
    --    Tms.Assert( Device.Stat = 0, "Device Status is OK (Expected 0," &
    --        " actual " & Integer'Image(Device.Stat) & ")");
    --
    -- Usage example:
    --
    --    Tms.Default_Log_Mode := Tms.Log_All;
    --
    --    Retval := Device.Status_Register1;
    --    Tms.Assert( Retval = 0, "Ensure device status reg1 is zero");
    --    Retval := Device.Status_Register2;
    --    Tms.Assert( Retval /= 0, "Ensure device status reg2 is non-zero");
    --


    function Check_Assertions return Assertion_Status;
    --
    -- Returns None if no assertions were made during this test case,
    -- All_True if all assertions evaluated to true, and Some_False if
    -- at least one assertion failed.  This function can be used to
    -- explicitly convert (previous) assertions to a test result.
    --
    -- Usage example:
    --
    --    procedure My_Eval_Action is
    --    begin
    --        if Tms.Check_Assertions = Tms.Some_False then
    --            Tms.Register_Result(Tms.Failed_Assertion);
    --        else
    --            Tms.Put_Line("Assertions OK; checking temperature");
    --            if Temperature.OK then
    --                Tms.Register_Result( Tms.Pass );
    --            else
    --                Tms.Register_Result( Tms.Fail );
    --            end if;
    --        end if;
    --    end My_Eval_Action;
    --


    generic
        with procedure Visit (Description : String; Value : Boolean);
    procedure Traverse (Only_Visit_Failures : Boolean := False);
    --
    -- Iterates through all assertions made during the current test
    -- case.  Intended usage is within the test case evaluation action.
    -- This function can be used to evaluate a test case based on the
    -- number of assertions.  (E.g., "test case passes if 75% of the
    -- assertions are True")
    --



    ------------------------------------------------------------------------
    --
    -- EXECUTION LOG
    --
    -- These routines allow explicit access to the test results log.
    --
    -- The test message log can be redirected independently of the
    -- test's Standard.Output stream, allowing you to separate informational
    -- messages from the output of the unit or program under test.
    --
    -- The TestMate switch TM_OUTPUT_FILTER controls which message kinds
    -- are entered into the log.  The TestMate switch TM_LOG_TIMESTAMPS
    -- controls whether or not the logged messages are prefixed with
    -- a timestamp.
    --
    type Log_Msg_Kind is (Nill,          -- no message generated
                          Auxiliary_Msg, -- ::: preceding message
                          Debug_Msg,     -- ??? preceding message
                          Note_Msg,      -- --- preceding message
                          Positive_Msg,  -- +++ preceding message
                          Position_Msg,  -- >>> preceding message
                          Negative_Msg,  -- ++* preceding message
                          Warning_Msg,   -- !!! preceding message
                          Error_Msg,     -- *** preceding message
                          Sharp_Msg,     -- ### preceding message
                          Dollar_Msg,    -- $$$ preceding message
                          At_Msg,        -- @@@ preceding message
                          Text_Msg);     --     unadorned message

    procedure Put_Line (Message : String; Kind : Log_Msg_Kind := Text_Msg);
    --
    -- Enters Message into the test log.
    --
    -- Usage example:
    --
    --    Tms.Put_Line("Starting test driver", Tms.Note_Msg);
    --
    -- Usage example:
    --
    --    Tms.Put_Line("Could not setup unit for testing", Tms.Error_Msg);
    --    Tms.Register_Result( Tms.Fail );
    --

    type Log_Filter is array (Log_Msg_Kind) of Boolean;
    --
    -- Specifies what types of messages are to printed.   A true value
    -- indicates that the associated message kind WILL be displayed; a
    -- false value indicates that the message will be filtered out.
    --

    procedure Set_Log_Filter (Filter : Log_Filter);
    --
    -- Sets the default log filter to be used for all subsequent calls
    -- to Put_Line.  This over-rides the setting of the TM_OUTPUT_FILTER
    -- switch.
    --


    ------------------------------------------------------------------------
    --
    -- Placeholder resolution
    --

    function Resolve_Placeholders (In_The_String : String) return String;
    --
    -- Allows user's code to resolve placeholder specifiers, e.g.
    -- <test_case:t:r>.  For a complete description of placeholder syntax,
    -- and for a list of all placeholders and modifiers, see the User's Guide
    --
    -- Usage example:
    --
    --    declare
    --        User : constant string := Tms.Resolve_Placeholders("<user>");
    --


    ------------------------------------------------------------------------
    --
    -- Access to switches (from switches files, command line, test list)
    --

    function Value (Of_The_Switch : String) return String;
    --
    -- Returns the value of a switch.  All switches are interpreted as
    -- strings; conversion to boolean, integer, and enumeration
    -- switch types is not performed.
    --
    -- See the user's guide for a list of all TestMate switches.
    -- You may also run the dialog TestMate:Reports:Switches, or
    -- run "testmate show_switches" from the command line.
    --
    -- Usage example:
    --
    --     Input_Loc : constant string := Tms.Value("TM_INPUT_LOC");
    --


    ------------------------------------------------------------------------
    --
    -- Access to various run-time values of the current test's results
    --

    function Number_Of_Tests (With_Result : Test_Result) return Natural;
    --
    -- Returns the number of tests in the current test list that have
    -- completed with the specified Test_Result kind.  Intended usage is
    -- from within evaluation actions or test list epilogs.
    --


    function Prior_Test_Case_Result return Test_Result;
    --
    -- Returns the result of the last test case to complete processing.
    -- Intended usage is from within test case preconditions to gate
    -- whether or not to run a test case.
    --
    -- Usage example:
    --
    --     function My_Precondition return Boolean is
    --     begin
    --         return Tms.Is_Equal( Tms.Prior_Test_Case_Result, Tms.Pass);
    --     end My_Precondition;
    --


    function The_Test_Case_Result (Of_Test_Case : String) return Test_Result;
    --
    -- Returns the test result of the currently active test case.  Note
    -- that if no result has yet been registered, this will return
    -- Unknown.
    --
    -- Note that Of_Test_Case must be part of the currently executing
    -- test list, and must been run prior to this call.
    --


    function Is_Equal (L, R : Test_Result) return Boolean;
    --
    -- This is a case insensitive equal.
    --
    -- Example:
    --
    --    if Tms.Is_Equal( Tms.Prior_Test_Case_Result, Tms.Pass ) then
    --      ...
    --



    ------------------------------------------------------------------------
    --
    -- Exporting UNIX environment to subsequent executables
    --

    procedure Register_Environment;
    --
    -- If environment variables need to be passed from one executable
    -- process to another during a test run (e.g. from the test list prolog
    -- to all test drivers), then this routine needs to be called from
    -- within the exporter, before exiting.  In the example just given,
    -- the test list prolog would call this routine prior to exiting,
    -- and the environment variables would be exported for use later
    -- by the test drivers.
    --


    function Connected return Boolean;
    --
    -- Tests whether this process is connected to TestMate's test execution
    -- server via a socket.  If the test was started by TestMate, then the
    -- server should be connected; if not, then Tms functionality is
    -- greatly restricted.
    --

    ------------------------------------------------------------------------
    --
    -- IMAGE FUNCTIONS
    --
    -- These functions produce a string from predefined types that do
    -- not always have image functions.  These can be used to print
    -- expected vs. actual results or to log values for debugging.
    --

    function Image (X : System.Address) return String;
    function Image (X : Float) return String;
    function Image (X : Long_Float) return String;
    function Image (X : Short_Float) return String;
    function Image (X : String) return String;
    function Image (X : Calendar.Time) return String;

    function Hex_Image (X : Integer) return String;
    function Hex_Image (X : Long_Integer) return String;


    Module_Information : constant String :=
       "@(#) $Id: tms.1.ada,v 1.5 07/04/2002 21:56:22 gsmith Exp $\";

end Tms;
pragma Api (Testmate);

