--+--------------------------------------------------------------------------+
--| Copyright 1992 - 2002 Rational Software Corporation. All Rights Reserved.|
--| This software contains proprietary and confidential information of       |
--| Rational and its suppliers. Use, disclosure or reproduction is           |
--| prohibited without the prior express written consent of Rational.        |
--+--------------------------------------------------------------------------+
--

with Unchecked_Conversion;
with System;
with Tm_Mcdc_Runtime_Support;

package body Tm_Mcdc_Allocation is

    pragma Suppress (Storage_Check);
    pragma Optimize (Space);

    function Allocate_Bits (Size_In_Bits : Natural) return System.Address is
        Size_In_Bytes : constant Natural :=
           (Size_In_Bits + System.Storage_Unit - 1) / System.Storage_Unit;
        Return_Value  : System.Address;
        use type System.Address;
    begin
        Return_Value := Tm_Mcdc_Runtime_Support.Allocator.Alloc (Size_In_Bytes);
        if Return_Value = System.No_Addr then
            Unfulfilled_Allocs := Unfulfilled_Allocs + 1;
        end if;
        return Return_Value;
    end Allocate_Bits;
    pragma Inline (Allocate_Bits);

    procedure Deallocate_Bits (Object       : System.Address;
                               Size_In_Bits : Natural) is
        Size_In_Bytes : constant Natural := (Size_In_Bits + 7) / 8;
    begin
        null;
    end Deallocate_Bits;
    pragma Inline (Deallocate_Bits);


    package body Heap is

        function As_Handle is new Unchecked_Conversion (System.Address, Handle);
        function As_Address is new Unchecked_Conversion
                                      (Handle, System.Address);

        function Array_Size (Length : Integer) return Integer is
            type Elements is array (Integer range 1 .. Length) of Element;
        begin
            return Elements'Size;
        end Array_Size;
        pragma Inline (Array_Size);

        function Allocate return Handle is
        begin
            return As_Handle (Allocate_Bits (Element'Size));
        end Allocate;

        function Allocate (Initial_Value : Element) return Handle is
            Result : Handle := Allocate;
        begin
            if Result /= null then
                Result.all := Initial_Value;
            end if;
            return Result;
        end Allocate;

        function Allocate_Array (Length : Integer) return Handle is
        begin
            return As_Handle (Allocate_Bits (Array_Size (Length)));
        end Allocate_Array;

        procedure Deallocate (The_Handle : in out Handle) is
        begin
            Deallocate_Bits (As_Address (The_Handle), Element'Size);
            The_Handle := null;
        end Deallocate;

        procedure Deallocate_Array (The_Handle : in out Handle;  
                                    Length     : Integer) is
        begin
            Deallocate_Bits (As_Address (The_Handle), Array_Size (Length));
            The_Handle := null;
        end Deallocate_Array;

    end Heap;


    package body Pool is

        function As_Handle is new Unchecked_Conversion (System.Address, Handle);
        function As_Address is new Unchecked_Conversion
                                      (Handle, System.Address);

        function Array_Size (Length : Integer) return Integer is
            type Elements is array (Integer range 1 .. Length) of Element;
        begin
            return Elements'Size;
        end Array_Size;
        pragma Inline (Array_Size);

        function Is_Pool_Handle (The_Handle : Handle; From : Pool.Data_Ptr)
                                return Boolean is
            function ">=" (L, R : System.Address) return Boolean
               renames System.">=";
            function "<" (L, R : System.Address) return Boolean
               renames System."<";
        begin
            return As_Address (The_Handle) >=
                      From.Storage (Pool_Index'First)'Address and then
                   As_Address (The_Handle) <
                      From.Storage (Pool_Index'Last)'Address;
        end Is_Pool_Handle;

        function Allocate (From : Pool.Data_Ptr) return Handle is
            Ix : Pool_Index := From.Next;
        begin
            if Ix < Pool_Index'Last then
                From.Next := Ix + 1;
                return As_Handle (From.Storage (Ix)'Address);
            end if;
            -- No more space in the data pool
            return As_Handle (Allocate_Bits (Element'Size));
        end Allocate;

        function Allocate (Initial_Value : Element;  
                           From          : Pool.Data_Ptr) return Handle is
            Result : Handle := Allocate (From);
        begin
            if Result /= null then
                Result.all := Initial_Value;
            end if;
            return Result;
        end Allocate;

        function Allocate_Array (Length : Integer;  
                                 From   : Pool.Data_Ptr) return Handle is
            Last_Ix : constant Pool_Index :=
               Pool_Index'Last - Pool_Index (Length);
            Ix      : Pool_Index          := From.Next;
        begin
            if Ix < Last_Ix then
                From.Next := Ix + Pool_Index (Length);
                return As_Handle (From.Storage (Ix)'Address);
            end if;
            -- No more space in the data pool
            return As_Handle (Allocate_Bits (Array_Size (Length)));
        end Allocate_Array;

        procedure Deallocate (The_Handle : in out Handle;  
                              From       : Pool.Data_Ptr) is
        begin
            if Is_Pool_Handle (The_Handle, From) then
                null; -- Do all deallocation in Clear.
            else
                Deallocate_Bits (As_Address (The_Handle), Element'Size);
            end if;
            The_Handle := null;
        end Deallocate;

        procedure Deallocate_Array (The_Handle : in out Handle;  
                                    Length     : Integer;  
                                    From       : Pool.Data_Ptr) is
        begin
            if Is_Pool_Handle (The_Handle, From) then
                null; -- Do all deallocation in Clear.
            else
                Deallocate_Bits (As_Address (The_Handle), Array_Size (Length));
            end if;
            The_Handle := null;
        end Deallocate_Array;

        procedure Clear (Pool_Data : Pool.Data_Ptr) is
        begin
            for Ix in Pool_Index'First .. Pool_Index'Last loop
                Pool_Data.Locks (Ix) := 0;
            end loop;
            Pool_Data.Next := Pool_Index'First;
        end Clear;
    end Pool;

end Tm_Mcdc_Allocation;
pragma Suppress_Elaboration_Checks;
