-- Copyright (C) 1995,2000,2002 Rational Software Corporation.
-- All rights reserved.

-- RESTRICTED RIGHTS LEGEND
--
-- Use, duplication, or disclosure by the Government and Boeing is subject to
-- restrictions as set forth in subparagraph (c)(1)(ii) of the Rights in
-- Technical Data and Computer Software clause of DFAR 252.227-7013 and in
-- similiar clauses in the FAR and NASA FAR Supplement.
------------------------------------------------------------------------------
-- PROPRIETARY AND CONFIDENTIAL INFORMATION OF RATIONAL SOFTWARE CORP.,
-- USE OR COPYING WITHOUT EXPRESS WRITTEN AUTHORIZATION IS STRICTLY PROHIBITED.
-- THIS MATERIAL IS PROTECTED AS AN UNPUBLISHED WORK UNDER THE U.S. COPYRIGHT
-- ACT OF 1976.  CREATED 1995.  ALL RIGHTS RESERVED.
------------------------------------------------------------------------------
--
-- File Name: tm_mcdc_score.1.ada
--
-- This package is the runtime interface package for MCDC instrumented
-- Ada source files.  All instrumented source references the subprograms
-- and data structures exported from this specification, and only these
-- declarations.  Each declaration listed in this package is followed by
-- an explanation of its use in instrumented code.
--
-- In general, there are three sections of subprograms:
--
-- 1.  Unit level registration declaration.
--
-- 2.  Decision marking routines, which come in two flavors:
--     a.  Bitmap relative decisions are the most common and simply set
--         a bit in the decision's units bitmap.
--     b.  Line relative decisions are either CASE statement expressions
--         or large (many condition) boolean expressions.
--
-- 3.  Condition bit setting routines.
--
-- The declarations available exist to support the above routines.  Each
-- declaration has a description that explains its intended use.
--

with Tm_Mcdc_Score_Support;
with System;

package Tm_Mcdc_Score is
    pragma Not_Elaborated;

    ---------------------------------------------------------------------
    -- Type Declarations
    ---------------------------------------------------------------------

    subtype Unit_Decisions is Tm_Mcdc_Score_Support.Unit_Decisions;
    --
    -- The is type used to allocate and access unit bitmaps.  Bitmaps
    -- are unit level arrays of booleans with an entry in the array for
    -- each possible decision result (that utilizes the unit bitmap).
    -- Not all decisions will utilize the bitmap, depending on whether
    -- the number of possible results is reasonable or not.
    --

    subtype Unit_Id is Tm_Mcdc_Score_Support.Unit_Id;
    --
    -- The type is a unique, dynamic identification for a unit.
    -- Some decisions use Unit_Ids to designate the unit in which they
    -- occur.
    --

    subtype Unit_Info is Tm_Mcdc_Score_Support.Unit_Info;
    --
    -- Used to make certain information local to a specific unit.
    -- ID is a unique dynamically allocated identifier that uniquely
    -- identifies a specific unit.  The ID is passed to some DECISION
    -- routines that do not utilize the unit BITMAP.  BITMAP is a
    -- list of potential decision results for a specific unit.  This
    -- list of boolean values is initialized to FALSE and each field
    -- is only set to TRUE when a particular decision result is
    -- dynamically encountered.
    --

    type Condition_Id is new Tm_Mcdc_Score_Support.Condition_Id;
    --
    -- A 32-bit "simple" bit map of condition values.  As
    -- subconditions are registered, the appropriate bit in a local
    -- CONDITION_ID is set to true (if the condition is true and
    -- left false if the condition is false).  The result once all
    -- conditions have been registered will be the decision result
    -- offset (from the decision bias) into the unit bitmap.
    --

    type Trace_Integer is new Tm_Mcdc_Score_Support.Trace_Integer;
    --
    -- An integer type that encompasses all possible values in the
    -- largest discrete type supported by the target.  This type is used
    -- to hold CASE statement results that have a potential span of
    -- min_int to max_int.
    --


    ---------------------------------------------------------------------
    -- Unit Registration Declarations
    ---------------------------------------------------------------------

    function Register (The_Unit                  : String;
                       The_Timestamp             : Integer;
                       The_Checksum              : Integer;
                       The_Decision_Array_Length : Natural;
                       The_Kind                  : Character := '#')
                      return Unit_Info;
    --
    -- Registers THE_UNIT with the runtime and returns a unique ID and
    -- access to an allocated decision BITMAP for this unit.  The returned
    -- UNIT_INFO is stored locally in each compilation unit, visible to
    -- each instrumented call inserted in THE_UNIT.  This allows direct
    -- access to the BITMAP and ID from anywhere within THE_UNIT.  The
    -- parameter THE_TIMESTAMP is simply additional information associated
    -- with THE_UNIT in the scorecard output.
    --


    ---------------------------------------------------------------------
    -- Decision Declarations
    ---------------------------------------------------------------------

    function Decision1 (The_Expr       : Boolean;
                        The_Unit_Array : Unit_Decisions;
                        The_Offset     : Condition_Id) return Boolean;
    --
    -- Mark the decision in the scorecard and return THE_EXPR value as
    -- the result of the call.  There is only one condition associated with
    -- THE_EXPR and its value is equal to THE_EXPR, so can use boolean'pos
    -- of THE_EXPR to get index from THE_OFFSET and no condition calls are
    -- used.
    --

    function Decision (The_Expr       : Boolean;
                       The_Unit_Array : Unit_Decisions;
                       The_Index      : System.Address;
                       The_Bias       : Condition_Id) return Boolean;
    --
    -- This call is used for multiple condition decisions that utilize the
    -- units BITMAP.  The bitmap is passed in as THE_UNIT_ARRAY and the
    -- appropriate offset into the array is computed based on the THE_BIAS
    -- plus THE_INDEX.  THE_BIAS is the offset into the units bitmap at
    -- which this decisions entries begin.  THE_INDEX is the offset into
    -- this decisions pure truth table to the appropriate decision result.
    --

    function Decisionx (The_Expr  : Boolean;
                        In_Unit   : Unit_Id;
                        The_Line  : Natural;
                        The_Value : System.Address) return Boolean;
    --
    -- Similar to the other DECISION routines, except that there are too
    -- many conditions associated with this decision to place it in the
    -- unit array and THE_VALUE must be cached independent of the normal
    -- unit bitmap. The mechanism used to cache these values is the same
    -- as the mechanism used to cache CASE expression results.
    --

    function Decisionx (The_Expr    : Boolean;
                        In_Unit     : Unit_Id;
                        The_Line    : Natural;
                        The_Line_Id : Natural;
                        The_Value   : System.Address) return Boolean;
    --
    -- Similar to the previous DECISIONX declaration, except that an
    -- additional THE_LINE_ID parameter is passed due to more then one
    -- decision being present on the same line.  The additional line id
    -- is necessary to guarantee uniqueness
    --

    function Decisionp (The_Expr    : Boolean;
                        In_Unit     : Unit_Id;
                        The_Line    : Natural;
                        The_Line_Id : Natural := 1) return Boolean;
    --
    -- Similar to the previous DECISIONX declaration, except that THE_VALUE
    -- is derived from THE_EXPR.
    --

    function Decisionx (The_Expr : Trace_Integer;
                        In_Unit  : Unit_Id;
                        The_Line : Natural) return Trace_Integer;
    --
    -- Similar to the other DECISIONX declarations this routine is used
    -- to cache THE_EXPR independent of the normal unit bitmap mechanism.
    -- This routine is used to monitor CASE statement expression results.
    -- Each CASE expression is type converted to TRACE_INTEGER passed to
    -- this routine which returns THE_EXPR which is then type converted
    -- back into the appropriate CASE expression type.
    --

    function Decisionx (The_Expr    : Trace_Integer;
                        In_Unit     : Unit_Id;
                        The_Line    : Natural;
                        The_Line_Id : Natural) return Trace_Integer;
    --
    -- Similar to the other DECISIONX declarations this routine is used
    -- to cache THE_EXPR independent of the normal unit bitmap mechanism.
    -- This routine is an auxiliary routine needed by support tools that
    -- generate scorecards independent of the normal runtime method.
    --

    function Decision
                (The_Unit_Array : Unit_Decisions; The_Offset : Condition_Id)
                return Boolean;
    --
    -- This routine simple sets "THE_UNIT_ARRAY (THE_OFFSET) := True" and
    -- returns a dummy true result.  This is used by subprogram entry
    -- monitoring to set the appropriate entry in the units bitmap before
    -- any declarations (with potential side-effects) are executed.
    --

    procedure Decision (The_Unit_Array : Unit_Decisions;
                        The_Offset     : Condition_Id);
    --
    -- This routine simply sets "THE_UNIT_ARRAY (THE_OFFSET) := True".
    -- This is used by subprogram entry & exit monitoring and FOR loop
    -- statement monitoring.  The appropriate entry in the units bitmap
    -- is set immediately before or after the appropriate statement in
    -- the source.
    --

    procedure Decision (The_Unit_Array : Unit_Decisions;
                        The_Offset     : Condition_Id;
                        The_Cond       : out Condition_Id);
    --
    -- This routine simply sets "THE_UNIT_ARRAY (THE_OFFSET) := True" and
    -- sets THE_COND to zero.  Used for complex FOR loop monitoring, this
    -- routine is used to clear the flag THE_COND after the true result
    -- has been set.
    --

    procedure Decision_If (The_Unit_Array : Unit_Decisions;
                           The_Offset     : Condition_Id;
                           The_Cond       : in out Condition_Id);
    --
    -- Set the boolean at THE_UNIT_ARRAY (THE_OFFSET) to TRUE if THE_COND
    -- equals -1.  Sets THE_COND to zero as a side-effect.  This routine
    -- is used to set the false result of a FOR loop construct in complex
    -- FOR constructs.
    --

    function Decision_Simple (The_Expr       : Boolean;
                              The_Unit_Array : Unit_Decisions;
                              The_Offset     : Condition_Id) return Boolean;
    --
    -- Set the boolean at THE_UNIT_ARRAY (THE_OFFSET) to TRUE and return
    -- THE_EXPR.
    --

    function Decision_Simple (The_Expr       : Boolean;
                              The_Unit_Array : Unit_Decisions;
                              The_Offset     : Condition_Id;
                              If_Equal       : Boolean) return Boolean;
    --
    -- Set the boolean at THE_UNIT_ARRAY (THE_OFFSET) to TRUE if THE_EXPR
    -- is equal to IF_EQUAL.  Always returns THE_EXPR.
    --


    ---------------------------------------------------------------------
    -- Condition Declarations
    ---------------------------------------------------------------------

    function Condition1
                (The_Expr : Boolean; The_Cond : System.Address) return Boolean;
    function Condition2
                (The_Expr : Boolean; The_Cond : System.Address) return Boolean;
    function Condition3
                (The_Expr : Boolean; The_Cond : System.Address) return Boolean;
    function Condition4
                (The_Expr : Boolean; The_Cond : System.Address) return Boolean;
    function Condition5
                (The_Expr : Boolean; The_Cond : System.Address) return Boolean;
    function Condition6
                (The_Expr : Boolean; The_Cond : System.Address) return Boolean;
    function Condition (The_Expr  : Boolean;
                        The_Cond  : System.Address;
                        The_Index : Condition_Id) return Boolean;
    --
    -- Set a specific bit in THE_COND (dereferenced) and returns THE_EXPR.
    -- The bit is determined by either the call number (1 thru 6) or
    -- THE_INDEX.  Note, that the CONDITION{1-6} routines perform the
    -- exact same functionality as the CONDITION routine with an implied
    -- value for THE_INDEX.  The use of multiple condition calls is used
    -- to reduce the amount of code space used by each condition call by
    -- one parameter.  If these routines are not inlined then the removal
    -- of the additional parameter becomes significant for multi-condition
    -- decision points in the program.
    --


private

    --
    -- This whole approach to instrumentation of the user code is based on
    -- utilization of inlining the condition calls to get the call size as
    -- small as possible.  Without inlining there is an issue of code bloat
    -- in instrumented units.  Functionally the inlining is not necessary,
    -- but with the actual code size of these routines being very small the
    -- inlining will aid significantly in overall code size expansion.
    --
    pragma Inline (Condition1);
    pragma Inline (Condition2);
    pragma Inline (Condition3);
    pragma Inline (Condition4);
    pragma Inline (Condition5);
    pragma Inline (Condition6);
    pragma Inline (Condition);

    pragma Inline (Decision1);
    pragma Inline (Decision);
    pragma Inline (Decision_Simple);
    --@@@ pragma Inline (Decisionx);

    --@@@ These routines should not be inlined; they are more verbose then
    --@@@ the call to the routines and are not as common as the others.

end Tm_Mcdc_Score;
pragma Api (Testmate);
