-- Copyright (C) 1995-1997,2000,2002 Rational Software Corporation.
-- All rights reserved.
--
------------------------------------------------------------------------------
-- RESTRICTED RIGHTS LEGEND
--
-- Use, duplication, or disclosure by the Government and Boeing is subject to
-- restrictions as set forth in subparagraph (c)(1)(ii) of the Rights in
-- Technical Data and Computer Software clause of DFAR 252.227-7013 and in
-- similiar clauses in the FAR and NASA FAR Supplement.
------------------------------------------------------------------------------
-- PROPRIETARY AND CONFIDENTIAL INFORMATION OF RATIONAL SOFTWARE CORP.,
-- USE OR COPYING WITHOUT EXPRESS WRITTEN AUTHORIZATION IS STRICTLY PROHIBITED.
-- THIS MATERIAL IS PROTECTED AS AN UNPUBLISHED WORK UNDER THE U.S. COPYRIGHT
-- ACT OF 1976.  CREATED 1995.  ALL RIGHTS RESERVED.
------------------------------------------------------------------------------
--
-- Map UNIT and LINE information to RESULT values.  This mapping is used to
-- build scorecard information from trace results.  The RESULT values are
-- collected together and returned as a single catenated result used to build
-- a scorecard.
--

with System;
with Unchecked_Conversion;
with Tm_Mcdc_Config;

package Tm_Mcdc_Score_Support is
    pragma Not_Elaborated;

    type Trace_Integer is range System.Min_Int .. System.Max_Int;

    type Condition_Id is range -1 .. +(2 ** 31) - 1;

    type Bool_Decision_Array is array (Condition_Id range <>) of Boolean;
    pragma Pack (Bool_Decision_Array);
    pragma Convention (Ada_Pass_By_Reference, Bool_Decision_Array);
    pragma Suppress (Index_Check, On => Bool_Decision_Array);

    type Unit_Decisions is
       access Bool_Decision_Array
                 (0 .. Tm_Mcdc_Config.Maximum_Number_Of_Conditions);
    for Unit_Decisions'Storage_Size use 0;
    function Decisions is new Unchecked_Conversion
                                 (System.Address, Unit_Decisions);

    subtype Unit_Id is System.Address;
    -- The address of a string with the unit's name

    Null_Unit_Id : constant Unit_Id := System.Null_Address;
    function "=" (L, R : Unit_Id) return Boolean renames System."=";

    type Unit_Info is
        record
            Id     : Unit_Id;
            Bitmap : Unit_Decisions;
        end record;
    pragma Convention (Ada_Pass_By_Reference, Unit_Info);

    Null_Unit_Info : constant Unit_Info := Unit_Info'(Null_Unit_Id, null);

    package Result_List is
        type Node is private;
        type List is access Node;
        for List'Storage_Size use 0;

        function Make (The_Result : Trace_Integer) return List;
        -- Create a new list, one element long containing only THE_RESULT.

        procedure Discard (The_List : in out List);
        -- Deallocate all memebers of THE_LIST.

        procedure Insert (On_List : List; The_Result : Trace_Integer);
        -- Attaches THE_RESULT to the end of ON_LIST if it is not a member
        -- of ON_LIST already.  ON_LIST must not be empty.

        generic
            with procedure Process_Element (The_Element : Trace_Integer);
        procedure Iterate (The_List : List);
        -- Walk THE_LIST and call PROCESS_ELEMENT for each element of the list.

        function Image (The_List : List) return String;
        -- Returns the catenation of every element in the list.  No separators
        -- are placed between the elements of THE_LIST.

        procedure Clear_Data_Pool;

    private
        type Node is
            record
                The_Result : Trace_Integer;
                The_Link   : List;
            end record;
        pragma Convention (Ada_Pass_By_Reference, Node);
    end Result_List;

    package Unit_Id_Map is
        type Node is private;
        type Set is access Node;
        for Set'Storage_Size use 0;

        type Map is array (Natural range 0 .. 63) of Set;
        Empty_Map : constant Map := Map'(others => null);

        procedure Destructor (The_Map : in out Map);
        -- Deallocate all elements of the map and return an empty map.

        procedure Find (In_Map          : Map;
                        The_Unit        : Unit_Id;
                        The_Line        : Natural;
                        The_Line_Id     : Natural;
                        The_Result_List : out Result_List.List;
                        Success         : out Boolean);
        -- Check IN_MAP to see if THE_UNIT+THE_LINE is a member of the map.
        -- If already a member SUCCESS is set to true and THE_RESULT_LIST is
        -- set appropriately.

        procedure Insert (Into_Map        : in out Map;
                          The_Unit        : Unit_Id;
                          The_Line        : Natural;
                          The_Line_Id     : Natural;
                          The_Result_List : Result_List.List);
        -- Add THE_UNIT+THE_LINE to INTO_MAP with THE_RESULT_LIST associated
        -- with the entry.

        generic
            with procedure Process_Element (The_Unit        : Unit_Id;
                                            The_Line        : Natural;
                                            The_Line_Id     : Natural;
                                            The_Result_List : Result_List.List);
        procedure Iterate (The_Map : Map);
        -- Walk the THE_MAP and call PROCESS_ELEMENT for each entry in the map.

    private
        type Node is
            record
                The_Unit        : Unit_Id;
                The_Line        : Natural;
                The_Line_Id     : Natural;
                The_Result_List : Result_List.List;
                The_Link        : Set;
            end record;

    end Unit_Id_Map;

    package Unit_Map is
        type Node;
        type Set is access Node;
        for Set'Storage_Size use 0;

        function Is_Null (L : Set; R : Set := null) return Boolean renames "=";

        type Node is
            record
                The_Id_Length        : Natural;
                The_Id               : Unit_Id;
                The_Timestamp        : Integer;
                The_Checksum         : Integer;
                The_Decisions_Length : Natural;
                The_Decisions        : Unit_Decisions;
                The_Kind             : Character;
                The_Link             : Set;
            end record;

        type Map is array (Natural range 0 .. 63) of Set;
        Empty_Map : constant Map := Map'(others => null);

        procedure Destructor (The_Map : in out Map);
        -- Deallocate all elements of the map and return an empty map.

        procedure Find (In_Map : Map; The_Unit : String; The_Set : out Set);
        -- Check IN_MAP to see if THE_UNIT is a member of the map.
        -- If already a member THE_SET designates the map element found,
        -- otherwise THE_SET is NULL.

        procedure Insert (Into_Map             : in out Map;
                          The_Unit             : String;
                          The_Timestamp        : Integer;
                          The_Checksum         : Integer;
                          The_Decisions_Length : Natural;
                          The_Decisions        : Unit_Decisions;
                          The_Kind             : Character;
                          The_Set              : out Set);
        -- Add THE_UNIT to INTO_MAP and set its attributes according to the
        -- parameters. If successful, THE_SET designates the newly created
        -- map element.

        procedure Replace (The_Set              : Set;
                           The_Unit             : String;
                           The_Timestamp        : Integer;
                           The_Checksum         : Integer;
                           The_Decisions_Length : Natural;
                           The_Decisions        : Unit_Decisions;
                           The_Kind             : Character);
        -- Update the attribute valus of THE_SET.

        generic
            with procedure Process_Element (The_Unit : String; The_Set : Set);
        procedure Iterate (The_Map : Map);
        -- Walk the THE_MAP and call PROCESS_ELEMENT for each entry in the map.

        function Is_Empty (The_Map : Map) return Boolean;
        -- Returns TRUE if THE_MAP = (others => null);

    end Unit_Map;

    type Scorecard is
        record
            The_Unit_Map    : Unit_Map.Map;
            The_Unit_Id_Map : Unit_Id_Map.Map;
        end record;
    pragma Convention (Ada_Pass_By_Reference, Scorecard);
    
    Nil_Scorecard : constant Scorecard :=
       (Unit_Map.Map'(others => null), Unit_Id_Map.Map'(others => null));

    procedure Destroy (The_Scorecard : in out Scorecard);
    --
    -- Deallocate all data associated with THE_SCORE and set THE_SCORE to
    -- NIL_SCORECARD.
    --

    type Scorecard_Ref is access Scorecard;
    for Scorecard_Ref'Storage_Size use 0;
    function Reference is new Unchecked_Conversion
                                 (System.Address, Scorecard_Ref);

    function Register (The_Score                 : Scorecard_Ref; -- in out
                       The_Unit                  : String;
                       The_Timestamp             : Integer;
                       The_Checksum              : Integer;
                       The_Decision_Array_Length : Natural;
                       The_Kind                  : Character) return Unit_Info;
    --
    -- Returns the unique id associated with THE_UNIT.  If THE_UNIT has not
    -- been seen earlier then the unique id is created and returned.
    --

    procedure Register_Decision (The_Score   : Scorecard_Ref; -- in out
                                 The_Unit    : Unit_Id;
                                 The_Line    : Natural;
                                 The_Line_Id : Natural;
                                 The_Result  : Trace_Integer);
    --
    -- Add THE_RESULT to THE_UNIT+THE_LINE entry in the THE_SCORE.  If
    -- THE_UNIT or THE_LINE are not present in the scorecard map they are
    -- inserted dynamically.
    --

    procedure Clean_Scorecard (The_Scorecard : in out Scorecard);
    --
    -- Clear all buffered dynamic coverage information from THE_SCORECARD,
    -- but does NOT remove any units registered in this scorecard.
    --

    function Is_Clean_Scorecard (The_Scorecard : Scorecard) return Boolean;
    --
    -- Is there any new data associated with THE_SCORECARD?  If yes then
    -- return FALSE, otherwise return TRUE.
    --

    generic
        with procedure Process_Unit (The_Unit           : String;
                                     The_Id             : Unit_Id;
                                     The_Timestamp      : Integer;
                                     The_Checksum       : Integer;
                                     The_Decision_Array : Bool_Decision_Array;
                                     The_Kind           : Character);
        with procedure Process_Entry (The_Unit_Id : Unit_Id;
                                      The_Line    : Natural;
                                      The_Line_Id : Natural;
                                      The_Result  : String);
    procedure Iterate (The_Score : Scorecard_Ref);
    --
    -- Iterates over the contents of the THE_SCORE scorecard and calls the
    -- supplied subprogram PROCESS_ENTRY for each THE_UNIT+THE_LINE combination
    -- in the scorecard.  Note, the returned THE_RESULT is a catenation of all
    -- registered results for THE_UNIT+THE_LINE.
    --

    generic
        with procedure Process_Statistic (The_Stat : String);
    procedure Dump_Statistics;
    --
    -- Calls PROCESS_STATISTIC for each monitored event in the MCDC runtime
    -- support packages.
    --

    generic
        with procedure Process_Warning (Warn_Msg : String);
    procedure Dump_Warnings;
    --
    -- Calls PROCESS_WARNING for each warning message
    --

    function Convert (The_Value : Bool_Decision_Array) return String;
    function Convert (The_Value : String) return Bool_Decision_Array;
    --
    -- Consolidate DECISION_ARRAY conversions to this one location so that
    -- they are performed in a consistent fashion.
    --

end Tm_Mcdc_Score_Support;
pragma Api (Testmate);
