--+--------------------------------------------------------------------------+
--| Copyright 1992 - 2004 Rational Software Corporation. All Rights Reserved.|
--| This software contains proprietary and confidential information of       |
--| Rational and its suppliers. Use, disclosure or reproduction is           |
--| prohibited without the prior express written consent of Rational.        |
--+--------------------------------------------------------------------------+
--
--[internal] Ada/LRM
with Text_Io;
with Unchecked_Deallocation;
--[internal]
--[internal] TestMate
with Tm_Mcdc_Score_Manager;
with Tms;
with Tms_Communications;
with Tms_Communications_Client;
with Tms_Script_Commands;
--[internal]
--[internal] Apex
with Apex.Debug_Tools;
with Apex.Unix_Types;  
with Apex.Unix_Operations;

package body Tms_Script_Support is

    package Client renames Tms_Communications_Client;
    package Sc renames Tms_Script_Commands;

    ------------------------------------------------------------------------
    -- Coverage Analyzer Support 
    ------------------------------------------------------------------------

    type User_Event is new Integer;
    Start_Tc : constant User_Event := 1;
    End_Tc   : constant User_Event := 2;

    procedure Trigger_User_Event
                 (The_User_Event : User_Event; The_Test_Case : Integer);
    --
    -- This procedure causes all Active events associated with
    -- THE_USER_EVENT to be processed (have their associated actions
    -- called).  Expected Ada client usage will be via the following
    -- pragmas in the user's code.
    --
    pragma Import (C, Trigger_User_Event,
                   External_Name => "TRIGGER_USER_EVENT");

    ------------------------------------------------------------------------
    -- Internal 
    ------------------------------------------------------------------------

    type String_Handle is new Sc.String_Handle;
    type Seconds is new Sc.Seconds;

    type Tc_Termination_Reason is new Sc.Tc_Termination_Reason;
    type Termination_Reason is new Sc.Termination_Reason;

    type Tl_Fields is
        record
            Overall_Number_Of_Failures : Natural := 0;
        end record;
    pragma Convention (Ada_Pass_By_Reference, Tl_Fields);

    type Tc_Fields is
        record
            Input                   : String_Handle         := null;
            Output                  : String_Handle         := null;
            Error_Output            : String_Handle         := null;
            Execution_Context       : String_Handle         := null;
            Allow_Coverage_Analysis : Boolean               := True;
            Cpu_Limit               : Seconds               := 0;
            Elapsed_Limit           : Seconds               := 0;
            Term_Reason             : Tc_Termination_Reason := None;
        end record;
    pragma Convention (Ada_Pass_By_Reference, Tc_Fields);

    type States is (Input_State, Output_State, Err_Output_State, Cwd_State);

    Tc_State : Tc_Fields;
    Tl_State : Tl_Fields;

    End_Test_Case_Reported : Boolean := False;

    Current  : array (States) of String_Handle := (others => null);
    Previous : array (States) of String_Handle := (others => null);

    procedure Trace (Msg : String) is
    begin
        Tms_Communications.Trace ("SCRIPT> " & Msg);
    end Trace;


    function Create (Str : String) return String_Handle is
    begin
        return new String'(Str);
    end Create;


    procedure Destroy is new Unchecked_Deallocation (String, String_Handle);


    function Value (Of_Handle : String_Handle) return String is
    begin
        if Of_Handle = null then
            return "";
        else
            return Of_Handle.all;
        end if;
    end Value;


    procedure Set_Tc_Term (Reason : Tc_Termination_Reason) is
    begin
        Tc_State.Term_Reason := Reason;
    end Set_Tc_Term;


    procedure Report_Tc_Term (Reason : Tc_Termination_Reason) is
    begin
        case Reason is
            when None =>
                --  test case completion will be reported by TES
                null;
            when Failed_Precondition =>
                Tms.Put_Line ("Precondition failed, driver skipped",
                              Tms.Negative_Msg);
            when Missing_Io_File =>
                Tms.Put_Line ("Error in test case input/output files",
                              Tms.Negative_Msg);
            when Elapsed_Time_Overflow =>
                Tms.Put_Line ("Test case exceeded elapsed time limit",
                              Tms.Negative_Msg);
            when Cpu_Time_Overflow =>
                Tms.Put_Line ("Test case exceeded CPU time limit",
                              Tms.Negative_Msg);
            when Unhandled_Exception =>
                Tms.Put_Line ("Test case received an unhandled exception",
                              Tms.Negative_Msg);
            when Internal_Error =>
                Tms.Put_Line ("Test case terminated due to an internal error",
                              Tms.At_Msg);
            when Other_Error =>
                Tms.Put_Line ("Test case terminated", Tms.Negative_Msg);
        end case;
    end Report_Tc_Term;


--[internal]    -- procedure Trace_Tc_Info (Header : String) is
--[internal]    -- begin
--[internal]    --     Trace (Header & " TC information:");
--[internal]    --     Trace ("  Execution_Context  = " & Value (Tc_State.Execution_Context));
--[internal]    --     Trace ("  Input              = " & Value (Tc_State.Input));
--[internal]    --     Trace ("  Output             = " & Value (Tc_State.Output));
--[internal]    --     Trace ("  Error_Output       = " & Value (Tc_State.Error_Output));
--[internal]    --     Trace ("  Allow_Coverage_Analysis = " &
--[internal]    --            Boolean'Image (Tc_State.Allow_Coverage_Analysis));
--[internal]    --     Trace ("  Cpu_Limit = " & Seconds'Image (Tc_State.Cpu_Limit));
--[internal]    --     Trace ("  Elapsed_Limit = " & Seconds'Image (Tc_State.Elapsed_Limit));
--[internal]    -- end Trace_Tc_Info;
--[internal]
--[internal]
    procedure Change_Cwd (To_Dir : String) is
        C_String_Dir : aliased constant String := To_Dir & Ascii.Nul; --[ada95]
--[ada83]        C_String_Dir : constant String := To_Dir & Ascii.Nul;
        Status       : Apex.Unix_Types.Int;
        --
        -- Should consider replacing these calls with POSIX
        --
        function Errno_Image return String is
        begin
            return Apex.Unix_Types.Int'Image (Apex.Unix_Operations.Get_Errno);
        end Errno_Image;

    begin
        Status := Apex.Unix_Operations.Chdir
                     (Apex.Unix_Types.To_Char_Ptr (C_String_Dir'Address));
        if not Apex.Unix_Types."=" (Status, 0) then
            Tms.Put_Line ("Could not chdir to execution context """ &
                          To_Dir & """ (errno = " & Errno_Image & ")",
                          Tms.Warning_Msg);
        end if;
    end Change_Cwd;

    ------------------------------------------------------------------------
    -- Exported subprograms
    ------------------------------------------------------------------------

    function Resolve_Placeholders (In_The_String : String) return String is
    begin
        return Sc.Resolve_Placeholders (In_The_String);
    end Resolve_Placeholders;


    procedure Initialize is
    begin
        -- First alert the CA that elaboration is done ...
        Trigger_User_Event (The_User_Event => End_Tc, The_Test_Case => 0);
        Tm_Mcdc_Score_Manager.End_Test_Case (0);

        -- Next let the TES know the script has initialized ...
        Tl_State.Overall_Number_Of_Failures := Sc.Initialize;
    exception
        when others =>
            -- message written in tms_script_commands
            Tl_State.Overall_Number_Of_Failures := 0;
            if not Client.Bastard then
                raise;
            end if;
    end Initialize;


    procedure Start_Test_Case is
        Term_Reason : Termination_Reason := Internal_Error;
    begin
        End_Test_Case_Reported := False;

        Sc.Start_Test_Case (Input_File     => Sc.String_Handle (Tc_State.Input),
                            Output_File    =>
                               Sc.String_Handle (Tc_State.Output),
                            Error_File     =>
                               Sc.String_Handle (Tc_State.Error_Output),
                            Exec_Context   =>
                               Sc.String_Handle (Tc_State.Execution_Context),
                            Allow_Coverage => Tc_State.Allow_Coverage_Analysis,
                            Cpu_Limit      => Sc.Seconds (Tc_State.Cpu_Limit),
                            Elapsed_Limit  =>
                               Sc.Seconds (Tc_State.Elapsed_Limit),
                            Term_Reason    =>
                               Sc.Termination_Reason (Term_Reason));

        Tc_State.Term_Reason := None;

        case Term_Reason is
            when None =>
                null; -- still okay
            when Elapsed_Time_Overflow =>
                Tms.Put_Line ("Script terminated due to Elapsed Time Overflow",
                              Tms.Error_Msg);
                raise Script_Halt;
            when Cpu_Time_Overflow =>
                Tms.Put_Line ("Script terminated due to CPU Time Overflow",
                              Tms.Error_Msg);
                raise Script_Halt;
            when Max_Number_Of_Failures_Exceeded =>
                Tms.Put_Line
                   ("Script terminated due to exceeding acceptable number of failures",
                    Tms.Error_Msg);
                raise Script_Halt;
            when Prior_Unhandled_Exception =>
                Tms.Put_Line
                   ("Script terminated due to unhandled exception in user test code",
                    Tms.Error_Msg);
                raise Script_Halt;
            when Internal_Error =>
                Tms.Put_Line ("Script terminated due to an internal error",
                              Tms.Error_Msg);
                raise Script_Halt;
            when Other_Error =>
                Tms.Put_Line ("Script terminated due to an error",
                              Tms.Error_Msg);
                raise Script_Halt;
        end case;

    exception
        when others =>
            if Client.Bastard then
                Tc_State.Allow_Coverage_Analysis := False;
                Tc_State.Cpu_Limit               := 0;
                Tc_State.Elapsed_Limit           := 0;
                Term_Reason                      := None;
            else
                raise;
            end if;
    end Start_Test_Case;


    procedure Start_Driver is
    begin
        Sc.Start_Driver;
    end Start_Driver;


    procedure End_Driver is
    begin
        Sc.End_Driver;
    end End_Driver;


    procedure End_Test_Case is
    begin
        if not End_Test_Case_Reported then
            End_Test_Case_Reported := True;
            Report_Tc_Term (Tc_State.Term_Reason);
            Sc.End_Test_Case (Sc.Tc_Termination_Reason (Tc_State.Term_Reason));
        end if;
    end End_Test_Case;


    procedure Test_Invoke (Id : Test_Case_Ids) is

        Encapsulated : Boolean := False;

        Early_Term : exception;

        Input_File  : Text_Io.File_Type;
        Output_File : Text_Io.File_Type;

        -----------------------------------------------------------------------

        procedure Set_Cwd (To_Dir : String_Handle) is
        begin
            Trace ("Set_Cwd entry");
            if Value (To_Dir) /= "" then
                Change_Cwd (Value (To_Dir));
                Current (Cwd_State) := Create (Value (To_Dir));
            end if;
        end Set_Cwd;


        procedure Reset_Cwd is
        begin
            if Current (Cwd_State) /= null then
                Change_Cwd (Value (Previous (Cwd_State)));
                Current (Cwd_State) := null;
            end if;
        end Reset_Cwd;

        procedure Set_Input (Filename : String) is
        begin
            if Filename = "" then
                Text_Io.Close (Input_File);
                Text_Io.Set_Input (Text_Io.Standard_Input);
            else
                Text_Io.Open (File => Input_File,
                              Mode => Text_Io.In_File,
                              Name => Filename);
                Text_Io.Set_Input (Input_File);
            end if;
        end Set_Input;

        procedure Set_Output (Filename : String) is
        begin
            if Filename = "" then
                Text_Io.Close (Output_File);
                Text_Io.Set_Output (Text_Io.Standard_Output);
            else
                begin
                    -- remove any existing output file
                    Text_Io.Open (File => Output_File,
                                  Mode => Text_Io.In_File,
                                  Name => Filename);
                    Text_Io.Delete (Output_File);
                exception
                    when others =>
                        null;
                end;
                Text_Io.Create
                   (File => Output_File,
                    Name => Filename,
                    Form =>
                       "OWNER => Read_Write, GROUP => Read_Write, OTHER => Read_Write");

                Text_Io.Set_Output (Output_File);
            end if;
        end Set_Output;

        procedure Set_Driver_State (To_Input       : String_Handle;
                                    And_Output     : String_Handle;
                                    And_Err_Output : String_Handle;
                                    Ok             : out Boolean) is
        begin
            Ok := True;
            Trace ("Set_State entry");
            if Value (To_Input) /= "" then
                Trace ("Set_State setting input to " & Value (To_Input));
                begin
                    Set_Input (Value (To_Input));
                    Current (Input_State) := Create (Value (To_Input));
                exception
                    -- when Io_Exceptions.Name_Error | Io_Exceptions.Mode_Error | 
                    --      Io_Exceptions.Use_Error | Io_Exceptions.Status_Error =>
                    when others =>
                        Tms.Put_Line
                           ("Could not redirect input from the file """ &
                            Value (To_Input) & """ (" &
                            Apex.Debug_Tools.Get_Exception_Name & ")",
                            Tms.Warning_Msg);
                        Ok := False;
                end;
            end if;

            if Value (And_Output) /= "" then
                begin
                    Trace ("Set_State setting output to " & Value (And_Output));
                    Set_Output (Value (And_Output));
                    Current (Output_State) := Create (Value (And_Output));
                exception
                    -- when Io_Exceptions.Name_Error | Io_Exceptions.Mode_Error | 
                    --      Io_Exceptions.Use_Error | Io_Exceptions.Status_Error =>
                    when others =>
                        Tms.Put_Line
                           ("Could not redirect output to the file """ &
                            Value (And_Output) & """ (" &
                            Apex.Debug_Tools.Get_Exception_Name & ")",
                            Tms.Warning_Msg);
                        Ok := False;
                end;
            end if;

            if Value (And_Err_Output) /= "" then
                Tms.Put_Line
                   ("Redirection of error output stream not supported",
                    Tms.Warning_Msg);
                Tms.Put_Line ("(attempted to redirect to " &
                              Value (And_Err_Output) & ')', Tms.Warning_Msg);
                -- begin
                --          Common_Text_Io.Create (File => Err_Output_File, 
                --                          Name => Value (And_Err_Output));
                --          -- check for exceptions...
                --          Common_Text_Io.Set_Error (Err_Output_File);
                --          Current (Err_Output_State) := Create (Value (And_Err_Output));
                -- exception
                --          when Io_Exceptions.Name_Error | Io_Exceptions.Mode_Error | 
                --               Io_Exceptions.Use_Error | Io_Exceptions.Status_Error =>
                --              pseudo_log ("Could not redirect stderr to the file """ & 
                --                         Value (And_Err_Output) & """ (" & 
                --                         Apex.Debug_Tools.Get_Exception_Name & ")");
                -- end;
            end if;

        end Set_Driver_State;


        procedure Reset_Driver_State is
        begin
            if Current (Input_State) /= null then
                Set_Input ("");
                Destroy (Current (Input_State));
                Current (Input_State) := null;
            end if;

            if Current (Output_State) /= null then
                Set_Output ("");
                Destroy (Current (Output_State));
                Current (Output_State) := null;
            end if;

            if Current (Err_Output_State) /= null then
                -- Text_Io.Close (File => Err_Output_File);
                -- -- check for exceptions...
                -- Text_Io.Set_Error (Text_Io.Standard_Error);
                -- Destroy (Current (Err_Output_State));
                Current (Err_Output_State) := null;
            end if;

        end Reset_Driver_State;

        -----------------------------------------------------------------------

        function Pre_Driver_Part return Boolean is
        begin
            Trace ("Test_Invoke.Pre_Driver_Part");
            begin
                Call_Prolog (Id);
            exception
                when others =>
                    Tms.Put_Line ("Unhandled exception in test case prolog: " &
                                  Apex.Debug_Tools.Get_Exception_Name
                                     (Machine_Info => True), Tms.Warning_Msg);
            end;
            begin
                return Call_Precondition (Id);
            exception
                when others =>
                    Tms.Put_Line
                       ("Unhandled exception in test case precondition: " &
                        Apex.Debug_Tools.Get_Exception_Name
                           (Machine_Info => True), Tms.Warning_Msg);
                    return False; --?
            end;
        end Pre_Driver_Part;

        procedure Driver_Part is
            Ok : Boolean := True;
        begin
            Trace ("Test_Invoke.Driver_Part");
            --
            -- make sure that the Start_Driver call occurs BEFORE
            -- the Set_Driver_State call.  The problem that I was
            -- seeing (defect 132539) is that somehow the socket
            -- interaction screws up the output file redirection...
            -- the symptom was that an output file of length zero
            -- was always created.  very strange.
            --
            Tms_Script_Support.Start_Driver;
            Set_Driver_State (Tc_State.Input, Tc_State.Output,
                              Tc_State.Error_Output, Ok);
            if Ok then
                begin
                    -- Next inform the CA that we are starting up the driver ...
                    if Tc_State.Allow_Coverage_Analysis then
                        Trigger_User_Event
                           (The_User_Event => Start_Tc,
                            The_Test_Case  => Test_Case_Ids'Pos (Id));
                        Tm_Mcdc_Score_Manager.Start_Test_Case;
                    end if;
                    Call_Test_Driver (Id);

                    -- After driver completes need to tell the CA about it ...
                    if Tc_State.Allow_Coverage_Analysis then
                        Trigger_User_Event
                           (The_User_Event => End_Tc,
                            The_Test_Case  => Test_Case_Ids'Pos (Id));
                        Tm_Mcdc_Score_Manager.End_Test_Case
                           (Test_Case_Ids'Pos (Id));
                    end if;
                    Tms_Script_Support.End_Driver;
                exception
                    when others =>
                        if Tc_State.Allow_Coverage_Analysis then
                            Trigger_User_Event
                               (The_User_Event => End_Tc,
                                The_Test_Case  => Test_Case_Ids'Pos (Id));
                            Tm_Mcdc_Score_Manager.End_Test_Case
                               (Test_Case_Ids'Pos (Id));
                        end if;
                        Tms.Put_Line
                           ("Unhandled exception in test case driver: " &
                            Apex.Debug_Tools.Get_Exception_Name
                               (Machine_Info => True), Tms.Warning_Msg);
                        Set_Tc_Term (Unhandled_Exception);
                        Reset_Driver_State;
                        raise Early_Term;
                end;
            else
                Set_Tc_Term (Missing_Io_File);
                raise Early_Term;
            end if;
            Reset_Driver_State;
        exception
            when others =>
                Reset_Driver_State;
                raise;
        end Driver_Part;


        procedure Post_Driver_Part is
        begin
            Trace ("Test_Invoke.Post_Driver_Part");
            begin
                Call_Evaluation_Action (Id);
            exception
                when others =>
                    Tms.Put_Line
                       ("Unhandled exception in test case evaluation action: " &
                        Apex.Debug_Tools.Get_Exception_Name
                           (Machine_Info => True), Tms.Warning_Msg);
            end;
            begin
                Call_Epilog (Id);
            exception
                when others =>
                    Tms.Put_Line ("Unhandled exception in test case epilog: " &
                                  Apex.Debug_Tools.Get_Exception_Name
                                     (Machine_Info => True), Tms.Warning_Msg);
            end;
        end Post_Driver_Part;


    begin
        Trace ("Entering Test_Invoke");
        Tms_Script_Support.Start_Test_Case;

        Set_Cwd (Tc_State.Execution_Context);

        Set_Tc_Term (None); -- assume success

        Encapsulated := False;
        if Pre_Driver_Part then
            Driver_Part;
            Post_Driver_Part;
        else
            Set_Tc_Term (Failed_Precondition);
            raise Early_Term;
        end if;

        Tms_Script_Support.End_Test_Case;

        Reset_Cwd;
        Trace ("Exiting Test_Invoke");
    exception
        when Early_Term =>
            Trace ("Test invoke: early termination");
            Tms_Script_Support.End_Test_Case;
            Reset_Cwd;
        when Script_Halt =>
            raise;  
        when others =>
            Tms.Put_Line ("Error invoking test case:  " &
                          Apex.Debug_Tools.Get_Exception_Name
                             (Machine_Info => True), Tms.Warning_Msg);
            Set_Tc_Term (Internal_Error);
            Tms_Script_Support.End_Test_Case;
            Reset_Cwd;
    end Test_Invoke;


end Tms_Script_Support;

